/* 
 *  DOCSIS configuration file encoder. 
 *  Copyright (c) 2001 Cornel Ciocirlan, ctrl@users.sourceforge.net.
 *  Copyright (c) 2002,2003 Evvolve Media SRL,office@evvolve.com
 *  
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *  DOCSIS is a registered trademark of Cablelabs, http://www.cablelabs.com
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <stdio.h>
#include <errno.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <fcntl.h>

#include "libdocsis.h"


void usage(char *prog_name ) {
	char VERSION[] = "0.3333333";

	printf( "DOCSIS Configuration File creator, version %s\n", VERSION); 
	printf ("Copyright (c) 1999,2000,2001 Cornel Ciocirlan, ctrl@users.sourceforge.net\n");
	printf ("Copyright (c) 2002,2003 Evvolve Media SRL, docsis@evvolve.com \n\n");
	printf ("To encode a cable modem configuration file: \n\t %s -e <modem_cfg_file> <key_file> <output_file>\n",prog_name);
	printf ("To encode a MTA configuration file: \n\t %s -p <mta_cfg_file> <output_file>\n",prog_name);
	printf ("To decode a CM or MTA config file: \n\t %s -d <binary_file>\n",prog_name);
	printf ("\nWhere:\n<modem_cfg_file>\t= name of text (human readable) configuration file\n<key_file>\t\t= text file containing the authentication key to be \n\t\t\t  used for the CMTS MIC (Message Integrity Check).\n<output_file> \t\t= name of output file where you want the binary data\n\t\t\t  to be written to (if it doesn't exist it is created).\n\t\t\t  This file can be TFTP-downloaded by DOCSIS-compliant\n\t\t\t  cable modems\n<binary_file>\t\t= name of binary file you to be decoded\n"); 
	printf ("\nSee examples/*.cfg for configuration file format.\n");
	printf ("\nPlease send bugs or questions to docsis-users@lists.sourceforge.net\n\n");
	exit ( -10 );
}

void decode_file(char *file);
void encode_file(char *infile, char *keyfile, char *outfile );


int main(int argc,char *argv[] ) {

	char		prog_name[256];
	char		err_buf[1024];
	FILE		*cf,*kf,*of;
	unsigned char	*buffer;
	unsigned int	buflen=0,keylen=0;
	unsigned int	is_mta = 0;
	int i;

	bzero( err_buf, 1024 );
	docsis_initialize( err_buf );

	memset (prog_name,0,255);
	strncpy (prog_name, argv[0], 254);

	if ( !(argc == 3 || argc == 4 || argc == 5 ) ) {
		usage (prog_name); 
		exit (10);
	}
	
	if ( !strcmp (argv[1],"-e")) {
		if ( argc != 5 ) usage(prog_name);
		if ( !strcmp (argv[2],argv[4]) ) {
			printf ("Error: source file is same as destination file\n");
			exit (-100); /* we don't overwrite the source file */
		}
		encode_file( argv[2], argv[3], argv[4] );
		if (*err_buf) fprintf( stderr, "%s\n", err_buf );
		exit(0);
	} else if ( !strcmp (argv[1], "-p")) {
		if ( argc != 4 ) usage(prog_name);
		if ( !strcmp (argv[2],argv[3]) ) { 
			printf ("Error: source file is same as destination file\n");
			exit (-100); 
		}
		encode_file( argv[2], NULL, argv[3] );
		if (*err_buf) fprintf( stderr, "%s\n", err_buf );
		exit(0);
	} else if ( !strcmp (argv[1],"-d")) {
		if (argc != 3 ) usage(prog_name);
		decode_file( argv[2] );
		if (*err_buf) fprintf( stderr, "%s\n", err_buf );
		exit(0);
	}
	/* no good options */
	usage(prog_name); 
}


void decode_file(char *file) {
	int ifd;
	unsigned char *buffer, *outbuf;
	unsigned int buflen=0, outlen;
	int rv=0;
	struct stat st;

	if ( (ifd = open (file, O_RDONLY))==-1 )  {
		printf("Error opening file %s: %s",file,strerror(errno));
		exit (-1);
	}	
	if ((rv=fstat(ifd, &st))) { 
		printf("stat on file %s: %s",file,strerror(errno));
		exit(-1);
	}
	
	buffer = (unsigned char *) calloc(1,st.st_size*sizeof(unsigned char) + 20);
	buflen = read(ifd, buffer, st.st_size);

	outbuf = (char *) calloc(1, 10240 );   /* 10 K should be good? */
	outlen = 10240;

	outlen = docsis_decode( buffer, buflen, outbuf, 10240 );

	printf("%s", outbuf );
}


int Read_KeyFile( char *keyfname, char *keybuffer ) {
	int		ifd, rv;
	struct stat	st;
	int		keylen;

	if ( (ifd = open (keyfname, O_RDONLY))==-1 )  {
		printf("Error opening file %s: %s",keyfname,strerror(errno));
		exit (-1);
	}	
	if ((rv=fstat(ifd, &st))) { 
		printf("stat on file %s: %s",keyfname,strerror(errno));
		exit(-1);
	}
	keylen = st.st_size;
	if (keylen < 1) {
		printf ("error: key must be at least 1 char long\n" );
		exit (-101);
	}
	if (keylen > 64) {
		printf ("error: key cannot be greater than 64 characters\n" );
		exit (-101);
	}
	keylen = read(ifd, keybuffer, st.st_size);

	while (keybuffer[keylen - 1] == 10 || keybuffer[keylen - 1] == 13) {
		keybuffer[keylen - 1] = 0;
		keylen--; /* eliminate trailing \n or \r */
	}
	return( keylen );
}


void encode_file(char *infile, char *keyfile, char *outfile ) {
	int		ifd, ofd, rv, keylen;
	struct stat	st;
	unsigned char	*buffer, *outbuf, *outb2;
	unsigned int	buflen=0, outlen, outl2;
	unsigned char	key[65];

	if ( (ifd = open (infile, O_RDONLY))==-1 )  {
		printf("Error opening file %s: %s",infile,strerror(errno));
		exit (-1);
	}	
	if ((rv=fstat(ifd, &st))) { 
		printf("stat on file %s: %s",infile,strerror(errno));
		exit(-1);
	}
	
	buffer = (unsigned char *) calloc(1, st.st_size*sizeof(unsigned char) + 20);
	buflen = read(ifd, buffer, st.st_size);
	close( ifd );

	// Since the encoder is kind of a compression technique the output
	// should be smaller than the input.
	outbuf = (char *) calloc(1, buflen*2 );   /* Double the input size */

	outlen = docsis_encode_whole( buffer, outbuf, buflen*2 );
	if (outlen == 0) {
		fprintf(stderr, "Error: Encoder returned no TLV data\n");
		return;
	}
	fprintf(stderr, "We received %d bytes of TLV data\n", outlen );
	if (keyfile != NULL) {
		keylen = Read_KeyFile( keyfile, key );
		/* Don't add CM MIC, CMTS MIC, EoD and pad unless it's a CM config file */
		outlen = docsis_add_cm_mic( outbuf, outlen, buflen*2 );
		outlen = docsis_add_cmts_mic( outbuf, outlen, buflen*2, key, keylen );
		outlen = docsis_add_eod_and_pad( outbuf, outlen, buflen*2 );
	}

	if ( (ofd = open( outfile, O_CREAT | O_WRONLY, 0444 ))==-1 )  {
		printf("Error opening file %s: %s",outfile,strerror(errno));
		exit (-1);
	}

	write( ofd, outbuf, outlen );
	close( ofd );
}

