/* 
 *  DOCSIS configuration file encoder. 
 *  Copyright (c) 2001 Cornel Ciocirlan, ctrl@users.sourceforge.net.
 *  Copyright (c) 2002 Evvolve Media SRL,office@evvolve.com
 *  
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *  DOCSIS is a registered trademark of Cablelabs, http://www.cablelabs.com
 */

#include <stdio.h>
#include <math.h>
#include <ctype.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>


#include "docsis_symtable.h"
#include "docsis_decode.h"
#include "docsis_encode.h"
#include "docsis_yy.tab.h"
#include "docsis_snmp.h"
#include "util_ethermac.h"


char *ob;  /* Output buffer */
int  omx;  /* Max Output buffer */
int   nt;  /* number of TABS to print */

// This prints out the number of tabs in global var 'nt'
#define prTAB				\
	{ int i;			\
	  for( i=0; i < nt; i++) {	\
	    sprintf(ob, "\t");		\
	    omx--; ob++;		\
	} }				\

#define OB  ob,omx

// This increments the output buffer
#define prVAL( a )		\
	{ snprintf a ;		\
	  omx -= strlen(ob);	\
	  ob += strlen(ob);	\
	}			\


void decode_uint (unsigned char *tlvbuf, symbol_type *sym) {
	unsigned int	helper = 0;

	memcpy( &helper, tlvbuf+2, sizeof(unsigned int) );
	prTAB; prVAL((OB, "%s\t%u;\n", sym->sym_ident, ntohl(helper) ));
}

void decode_ushort (unsigned char *tlvbuf, symbol_type *sym) {
	unsigned short	helper = 0;

	memcpy( &helper, tlvbuf+2, sizeof(unsigned short) );
	prTAB; prVAL((OB, "%s\t%hu;\n", sym->sym_ident, ntohs(helper) ));
}

void decode_uchar (unsigned char *tlvbuf, symbol_type *sym) {
	prTAB; prVAL((OB, "%s\t%hhu;\n", sym->sym_ident, (unsigned char) *(tlvbuf+2) ));
}

void decode_ip (unsigned char *tlvbuf, symbol_type *sym) {
	static struct in_addr	helper;

	memcpy (&helper, tlvbuf+2, sizeof(struct in_addr));
	prTAB; prVAL((OB, "%s\t%s;\n", sym->sym_ident, inet_ntoa(helper) ));
}

void decode_ether (unsigned char *tlvbuf, symbol_type *sym) {
	prTAB; prVAL((OB, "%s\t%s;\n", sym->sym_ident, ether_ntoa(tlvbuf+2) ));
}

void decode_ethermask (unsigned char *tlvbuf, symbol_type *sym) {
  // the return value of ether_ntoa is a pointer to a static string 
  // in the ether_ntoa function, so we have to print the values in two
  // "passees" to avoid the 2nd call overwriting the 1st. 

	prTAB;
	prVAL((OB, "%s\t%s/", sym->sym_ident, ether_ntoa(tlvbuf+2) ));
	prVAL((OB, "%s;\n", ether_ntoa(tlvbuf+8) ));
}

void decode_md5 (unsigned char *tlvbuf, symbol_type *sym) {
	int j;

	prTAB;
	prVAL((OB, "/* %s ", sym->sym_ident ));
	for( j=0; j < 16; j++ ) {
		prVAL((OB, "%02x", tlvbuf[j+2] ));
	}
	prVAL((OB, "; */\n" ));
}

void decode_snmp_wd (unsigned char *tlvbuf, symbol_type *sym) {
	prTAB;
	prVAL((OB, "%s ", sym->sym_ident ));

	/* last char in this TLV is not part of OID */
	decode_snmp_oid( tlvbuf+2, tlvbuf[1]-1, ob ); 
	omx -= strlen(ob); ob += strlen(ob);

	prVAL((OB, " %d ;\n", (unsigned int) tlvbuf[tlvbuf[1]+1] ));
}

void decode_oid (unsigned char *tlvbuf, symbol_type *sym) {
	prTAB;
	prVAL((OB, "%s ", sym->sym_ident ));
	decode_snmp_oid( tlvbuf+2, tlvbuf[1], ob );
	omx -= strlen(ob); ob += strlen(ob);
	prVAL((OB, ";\n" ));
}

void decode_snmp_object (unsigned char *tlvbuf, symbol_type *sym) {
	prTAB;
	prVAL((OB, "%s ", sym->sym_ident ));
	decode_vbind( tlvbuf+2, (unsigned int) tlvbuf[1], ob );
	omx -= strlen(ob); ob += strlen(ob);
/*  22-06-03  decode_vbind prints the trailing ';' as well  */
	prVAL((OB, "\n" ));
}

void decode_string (unsigned char *tlvbuf, symbol_type *sym) {
	char	*helper; 

	helper = (char *) alloca( ((unsigned int) tlvbuf[1])+1 );
	memset( helper, 0, tlvbuf[1]+1 );
	strncpy ( helper, (char *) tlvbuf+2, (unsigned int) tlvbuf[1] );
	prTAB; prVAL((OB, "%s \"%s\";\n", sym->sym_ident, helper ));
}

void decode_strzero (unsigned char *tlvbuf, symbol_type *sym) {
	char	*helper;

	helper = (char *) alloca( ((unsigned int) tlvbuf[1])+1 );
	memset( helper, 0, tlvbuf[1]+1);
	strncpy( helper, (char *) tlvbuf+2, (unsigned int) tlvbuf[1] );
	prTAB; prVAL((OB, "%s \"%s\";\n", sym->sym_ident, helper ));
}


void decode_hexstr (unsigned char *tlvbuf, symbol_type *sym) {
	char		*helper; 
	unsigned int	i;
	unsigned int	len;

	/* TODO  What? */
	len = (unsigned int) tlvbuf[1]; 
	helper = (char *) alloca( len+1 ); 
	memset( helper, 0, len+1 );
	memcpy( helper, (char *) tlvbuf+2, len );
	prTAB; prVAL((OB, "%s\t0x", sym->sym_ident ));
	for( i=0; i < len; i++ ) {
		prVAL((OB, "%02x", (unsigned char) helper[i] ));
	}
	prVAL((OB, ";\n" ));
}

void decode_unknown (unsigned char *tlvbuf, symbol_type *sym) {
	prTAB;
  	prVAL((OB, "GenericTLV\tTlvCode %d TlvLength %d TlvValue ",
			tlvbuf[0], tlvbuf[1] ));

	snprint_hexadecimal( ob,omx, (char *) tlvbuf+2, tlvbuf[1] );
	omx -= strlen(ob); ob += strlen(ob);

	if ( str_isprint( (char *) tlvbuf+2, tlvbuf[1] ) ) {
		prVAL((OB, "  /* %s */", tlvbuf[1], (char *) tlvbuf+2 ));
	}
	prVAL((OB, "\n" ));
}

void decode_special (unsigned char *tlvbuf, symbol_type *sym) {
	prVAL((OB, "%s\n", sym->sym_ident ));
}

void decode_aggregate( unsigned char *tlvbuf, symbol_type *sym ) {
	unsigned char	*cp;
	symbol_type	*current_symbol;
  
	cp = tlvbuf + 2; /* skip type,len of parent TLV */

	prTAB; prVAL((OB, "%s\n", sym->sym_ident ));
	prTAB; prVAL((OB, "{\n" ));
	nt++;	// Increase Indentation

	while ( (cp - tlvbuf) < (int) tlvbuf[1] ) {
		current_symbol = find_symbol_by_code_and_pid(cp[0], sym->id);
		if (current_symbol == NULL) {
			decode_unknown( cp, NULL );
		} else {
			current_symbol->decode_func (cp, current_symbol);
		}
		cp += cp[1] + 2;
	}

	nt--;	// Decrease Indentation
	prTAB; prVAL((OB, "}\n" ));
}

/* 
** This is just like decode_aggregate, but it only does symbol lookup 
** for the VendorIdentifier which is the only 'known' TLV.
**/

void decode_vspecific (unsigned char *tlvbuf, symbol_type *sym) {
	unsigned char	*cp;
	symbol_type	*first_symbol;

	cp = tlvbuf + 2; /* skip type,len of parent TLV */

	prTAB; prVAL((OB, "%s\n", sym->sym_ident ));
	prTAB; prVAL((OB, "{\n" ));
	nt++;	// Increase Indentation

	/* First TLV inside VendorSpecific has to be VendorIdentifier... the rest we don't care */ 
  
	first_symbol = find_symbol_by_code_and_pid( cp[0], sym->id );   
	if (first_symbol == NULL) {
		prTAB; prVAL((OB, "/* WARNING: Invalid VendorSpecific option \n" ));
		prTAB; prVAL((OB, "            1st element is NOT VendorIdentifier */\n" ));
		decode_unknown(cp, NULL);
	} else {
		/* Symbol found ... check if it's the right one */ 
		if (!strncmp (first_symbol->sym_ident, "vendoridentifier", 16)) {
			first_symbol->decode_func(cp, first_symbol);
		} else {
			prTAB; prVAL((OB, "/* Invalid VendorSpecific option \n" ));
			prTAB; prVAL((OB, "    1st element is NOT VendorIdentifier */\n" ));
			decode_unknown(cp, NULL);
		}
        }

	cp += cp[1] + 2;
	// cp = (unsigned char*) cp + (((unsigned char)*(cp+1))+2)*sizeof(unsigned char);
	// this piece of code tried to hurt my brain :)

	while ( (cp - tlvbuf) < (int) tlvbuf[1] ) {
		decode_unknown( cp, NULL );
		cp += cp[1] + 2;
	}

	nt--;	// Decrease Indentation
	prTAB; prVAL((OB, "}\n" ));
}



/* 
 * This function is needed because we don't have a symbol to call it. 
 * We can't put a "Main" symbol in the symtable because docsis_code is 
 * unsigned char (in symbol_type) and we reserve the values for 
 * DOCSIS use. 
 * It's also a bif different from docsis_aggregate in that docsis_aggregate 
 * takes an aggregate tlvbuf as argument that INCLUDES the "parent" code and 
 * length. On the main aggregate we don't have a code / length. 
 */

int decode_buffer(unsigned char *tlvbuf, int buflen,
		   char *outbuf, int maxbuf ) {
	unsigned char	*cp;
	symbol_type	*current_symbol;

	// Global Variables
	ob = outbuf;		// set up output buffer
	omx = maxbuf - 10;	// Just to be safe
  	nt = 0;

	cp = tlvbuf; 
	prVAL((OB, "Main \n{\n" ));
	nt++;	// Increase Indentation

	while ( (cp - tlvbuf) < buflen ) {
		current_symbol = find_symbol_by_code_and_pid (cp[0],0);
		if (current_symbol == NULL) { 
			decode_unknown(cp, NULL);	
		} else { 
			current_symbol->decode_func (cp, current_symbol);
		}
		cp += cp[1] + 2;
	}
	nt--;	// Decrease Indentation
	prVAL((OB, "}\n" ));
	return( ob - outbuf );
}


int str_isprint( const char *str, size_t str_len ) {
	int i;
	for(i=0; i < str_len; i++ )
		if (!(isprint((int) str[i]))) return 0;
	return 1;
}

