/* 
 *  DOCSIS configuration file encoder. 
 *  Copyright (c) 2001 Cornel Ciocirlan, ctrl@users.sourceforge.net.
 *  Copyright (c) 2002 Evvolve Media SRL,office@evvolve.com
 *  
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *  DOCSIS is a registered trademark of Cablelabs, http://www.cablelabs.com
 */

#include <stdio.h>
#include <errno.h>
#include <string.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <netinet/in.h>
#include <arpa/inet.h>

#include <fcntl.h>
#include <math.h>

#include "docsis_symtable.h"
#include "docsis_decode.h"
#include "docsis_encode.h"
#include "docsis_yy.tab.h"
#include "docsis_snmp.h"
#include "util_ethermac.h"

extern unsigned int line; 	/* defined in docsis_lex.l */
extern char *errmsg;


int encode_uint ( unsigned char *buf, void *tval, symbol_type *sym_ptr ) { 
	unsigned int	int_value;
	union t_val	*helper;

	if ( buf == NULL ) {
		sprintf( errmsg, "encode_uint called w/NULL buffer!\n");
		errmsg += strlen(errmsg);
		return 0;
	}

	if ( tval == NULL ) {
		sprintf( errmsg, "encode_uint called w/NULL value struct !\n");
		errmsg += strlen(errmsg);
		return 0;
	}

	helper = (union t_val *) tval;
	if ( sym_ptr->low_limit || sym_ptr->high_limit ) { 
		if ( helper->uintval < sym_ptr->low_limit || helper->uintval > sym_ptr->high_limit ) {
			sprintf( errmsg, "line %d, %s value %d out of range %hd-%hd\n ", 
				line, sym_ptr->sym_ident,
				helper->uintval,  sym_ptr->low_limit,
				sym_ptr->high_limit); 
			errmsg += strlen(errmsg);
			return 0;
		}
	}

	int_value  = htonl( helper->uintval );
#ifdef DEBUG
	sprintf( errmsg, "encode_uint: found %s value %d\n",sym_ptr->sym_ident, helper->uintval);
	errmsg += strlen(errmsg);
#endif /* DEBUG */
	memcpy ( buf, &int_value, sizeof(unsigned int)); 
	return( sizeof(unsigned int) );
}


int encode_ushort ( unsigned char *buf, void *tval, symbol_type *sym_ptr ) {
	unsigned short	sint;
	union t_val	*helper;

	if ( buf == NULL ) {
		sprintf( errmsg, "encode_ushort called w/NULL buffer!\n");
		errmsg += strlen(errmsg);
		return 0;
	}

	if ( tval == NULL  ) {
		sprintf( errmsg, "encode_ushort called w/NULL value struct !\n");
		errmsg += strlen(errmsg);
		return 0;
	}

	helper = (union t_val *) tval;
	if ( sym_ptr->low_limit || sym_ptr->high_limit ) { 
		if ( helper->uintval < sym_ptr->low_limit || helper->uintval > sym_ptr->high_limit ) {
			sprintf( errmsg, "line %d, %s value %d out of range %hd-%hd\n ",
				line,   sym_ptr->sym_ident,
				helper->uintval,  sym_ptr->low_limit,
				sym_ptr->high_limit); 
			errmsg += strlen(errmsg);
			return 0;
		}
	}

	sint = htons( (unsigned short) helper->uintval );
#ifdef DEBUG
	sprintf( errmsg, "encode_ushort: found %s value %hd\n",sym_ptr->sym_ident, helper->uintval);
	errmsg += strlen(errmsg);
#endif /* DEBUG */
	memcpy ( buf, &sint, sizeof(unsigned short) );
	return ( sizeof(unsigned short) );
}

int encode_uchar ( unsigned char *buf, void *tval, symbol_type *sym_ptr ) {
	unsigned int	int_value;
	char		*cp;
	union t_val	*helper;

	if ( buf == NULL ) {
		sprintf( errmsg, "encode_uchar called w/NULL buffer!\n");
		errmsg += strlen(errmsg);
		return 0;
	}

	if ( tval == NULL  ) {
		sprintf( errmsg, "encode_uchar called w/NULL value struct !\n");
		errmsg += strlen(errmsg);
		return 0;
	}

	helper = (union t_val *) tval;
	int_value = htonl( helper->uintval );

	if ( sym_ptr->low_limit || sym_ptr->high_limit ) { 
		if ( helper->uintval < sym_ptr->low_limit || helper->uintval > sym_ptr->high_limit ) {
			sprintf( errmsg, "line %d, %s value %d out of range %hd-%hd\n ",
				line,  sym_ptr->sym_ident,
				helper->uintval,  sym_ptr->low_limit,
				sym_ptr->high_limit ); 
			return 0;
		}
	}

	cp = (char *) &int_value;
	buf[0] = (unsigned char) cp[3];
#ifdef DEBUG
	sprintf( errmsg, "encode_uchar: found %s value %hd\n",sym_ptr->sym_ident, helper->uintval);
	errmsg += strlen(errmsg);
#endif 
	return ( sizeof(unsigned char) );
}             


int encode_ip( unsigned char *buf, void *tval, symbol_type *sym_ptr ) {
	struct in_addr	in;
	int		retval;
	union t_val	*helper;

	if ( buf == NULL ) {
		sprintf( errmsg, "encode_ip called w/NULL buffer!\n");
		errmsg += strlen(errmsg);
		return 0;
	}

	if ( tval == NULL  ) {
		sprintf( errmsg, "encode_ip called w/NULL value struct !\n");
		errmsg += strlen(errmsg);
		return 0;
	}

	helper = (union t_val *) tval;

	if (!(retval = inet_aton( helper->strval, &in) ) ) { 
		sprintf( errmsg, "Invalid IP address %s at line %d", helper->strval, line );
		errmsg += strlen(errmsg);
		return 0;
	}

#ifdef DEBUG
	sprintf( errmsg, "encode_ip: found %s at line %d\n", inet_ntoa(in), line);
	errmsg += strlen(errmsg);
#endif /* DEBUG */
	memcpy ( buf, &in, sizeof(struct in_addr) );
	return ( sizeof(struct in_addr) );
}


int encode_ether ( unsigned char *buf, void *tval, symbol_type *sym_ptr ) {
	int		retval;
	union t_val	*helper;

	if ( buf == NULL ) {
		sprintf( errmsg, "encode_ether called w/NULL buffer!\n");
		errmsg += strlen(errmsg);
		return 0;
	}

	if ( tval == NULL  ) {
		sprintf( errmsg, "encode_ether called w/NULL value struct !\n");
		errmsg += strlen(errmsg);
		return 0;
	}

	helper = (union t_val *) tval;

	if (!(retval = ether_aton( helper->strval, buf ) ) ) { 
		sprintf( errmsg, "Invalid MAC address %s at line %d", helper->strval, line );
		errmsg += strlen(errmsg);
		return 0;
	}

#ifdef DEBUG
	sprintf( errmsg, "encode_ether: found %s at line %d\n", ether_ntoa(buf), line );
	errmsg += strlen(errmsg);
#endif  /* DEBUG */
	return retval;  /* hopefully this equals 6 :) */     
}


int encode_ethermask ( unsigned char *buf, void *tval, symbol_type *sym_ptr ) {
	int		reta, retb;
	char		*ether,*mask;
	union t_val	*helper;

	if ( buf == NULL ) {
		sprintf( errmsg, "encode_ethermask called w/NULL buffer!\n");
		errmsg += strlen(errmsg);
		return 0;
	}

	if ( tval == NULL  ) {
		sprintf( errmsg, "encode_ethermask called w/NULL value struct !\n");
		errmsg += strlen(errmsg);
		return 0;
	}

	helper = (union t_val *) tval;
	ether = helper -> strval;

	mask = strchr( ether, (int) '/');
	if (mask == NULL) { 
		sprintf( errmsg, "encode_ethermask: at line %d, format should be <mac_address>/<mac_mask>\n", line);
		errmsg += strlen(errmsg);
		return 0;
	}

	mask[0]=0x0; mask++; /* cut the string in two */

	if (!(reta = ether_aton ( ether, buf)) ) { 
		sprintf( errmsg, "Invalid MAC address %s at line %d\n", ether, line );
		errmsg += strlen(errmsg);
		return 0;
	}
	if (!(retb = ether_aton ( mask, buf+reta*(sizeof(char))) ) ) {
		sprintf( errmsg, "Invalid MAC address %s at line %d\n", mask, line );
		errmsg += strlen(errmsg);
		return 0;
	}

#ifdef DEBUG
	sprintf( errmsg, "encode_ethermask: found %s/%s at line %d\n",
		ether_ntoa(buf), ether_ntoa(buf+reta*sizeof(char)), line);
	errmsg += strlen(errmsg);
#endif  /* DEBUG */
	return( reta + retb );  /* hopefully this equals 12 :) */     
}
 
int encode_string(unsigned char *buf, void *tval, symbol_type *sym_ptr ) { 
	unsigned int	string_size;
	union t_val	*helper; 

	if ( buf == NULL ) {
		sprintf( errmsg, "encode_string called w/NULL buffer!\n");
		errmsg += strlen(errmsg);
		return 0;
	}

	if ( tval == NULL  ) {
		sprintf( errmsg, "encode_string called w/NULL value struct !\n");
		errmsg += strlen(errmsg);
		return 0;
	}

	helper = (union t_val *) tval;
	string_size = strlen ( helper->strval );

	if (sym_ptr->low_limit || sym_ptr->high_limit) { 
		if ( string_size < sym_ptr->low_limit ) { 
			sprintf( errmsg, "encode_string: String too short, must be min %d chars\n", 
						sym_ptr->low_limit);
			errmsg += strlen(errmsg);
			return 0;
		}
		if ( sym_ptr->high_limit < string_size ) {
			sprintf( errmsg, "encode_string: String too long, must be max %d chars\n", 
							sym_ptr->high_limit);  
			errmsg += strlen(errmsg);
			return 0;
		}
	}

#ifdef DEBUG
	sprintf( errmsg, "encode_string: found '%s' on line %d\n", helper->strval, line );
	errmsg += strlen(errmsg);
#endif /* DEBUG */ 
	memset( buf, 0, string_size + 1 );
	memcpy( buf, helper->strval, string_size );
	return( string_size );  
}

/* This is for strings which need the terminating 0 at the end, e.g. Service Flow Class Name */

int encode_strzero(unsigned char *buf, void *tval, symbol_type *sym_ptr ) {
	unsigned int	string_size;
	union t_val	*helper;

	if ( buf == NULL ) {
		sprintf( errmsg, "encode_strzero called w/NULL buffer!\n");
		errmsg += strlen(errmsg);
		return 0;
	}

	if ( tval == NULL  ) {
		sprintf( errmsg, "encode_strzero called w/NULL value struct !\n");
		errmsg += strlen(errmsg);
		return 0;
	}

	helper = (union t_val *) tval;
	string_size = strlen ( helper->strval );
	if (sym_ptr->low_limit || sym_ptr->high_limit) {
		if ( string_size < sym_ptr->low_limit ) {
			sprintf( errmsg, "encode_string: String too short, must be min %d chars\n",
					sym_ptr->low_limit);
			errmsg += strlen(errmsg);
			return 0;
		}
		if ( sym_ptr->high_limit < string_size ) {
			sprintf( errmsg, "encode_string: String too long, must be max %d chars\n",
					sym_ptr->high_limit);
			errmsg += strlen(errmsg);
			return 0;
		}
	}

#ifdef DEBUG
	sprintf( errmsg, "encode_string: found '%s' on line %d\n", helper->strval, line );
	errmsg += strlen(errmsg);
#endif /* DEBUG */
	memset( buf, 0, string_size + 1 );
	memcpy( buf, helper->strval, string_size);
	return( string_size + 1 );
}

int encode_hexstr (unsigned char *buf, void *tval, symbol_type *sym_ptr) { 
	unsigned int	fragval;
	unsigned int	i;
	int		rval;
	char		*p;
	unsigned int	string_size;
	union t_val	*helper; 

	if ( buf == NULL ) {
		sprintf( errmsg, "encode_hexstr called w/NULL buffer!\n");
		errmsg += strlen(errmsg);
		return 0;
	}

	if ( tval == NULL  ) {
		sprintf( errmsg, "encode_hexstr called w/NULL value struct !\n");
		errmsg += strlen(errmsg);
		return 0;
	}

	helper = (union t_val *) tval;
	string_size = strlen ( helper->strval );

	if ( string_size != 2*floor(string_size/2) ) {
		sprintf( errmsg, "encode_hexstr: invalid hex string !\n");
		errmsg += strlen(errmsg);
		return 0;
	}

	p = helper->strval;

	if (p[0] != '0' || ( p[1] != 'x' && p[1] != 'X' )) {
		sprintf( errmsg, "encode_hexstr: invalid hex string %s\n", p);
		errmsg += strlen(errmsg);
		return 0;
	}

	p += 2*sizeof(char);
	i=0; 
	while (*p) { 
		if ( (rval=sscanf(p, "%02x", &fragval)) == 0) {
			sprintf( errmsg, "Invalid Hex Value %s\n", helper->strval);
			errmsg += strlen(errmsg);
			return 0;
		}
		buf[i] = (char) fragval; i++;
		p += 2*sizeof(char);
	}

	if (sym_ptr->low_limit || sym_ptr->high_limit) { 
		if ( i < sym_ptr->low_limit ) { 
			sprintf( errmsg, "encode_hexstr: Hex value too short, must be min %d octets\n", 
							sym_ptr->low_limit);
			errmsg += strlen(errmsg);
			return 0;
		}
		if ( sym_ptr->high_limit < i ) {
			sprintf( errmsg, "encode_hexstr: Hex value too long, must be max %d octets\n", 
							sym_ptr->high_limit);
			errmsg += strlen(errmsg);
			return 0;
		}
	}

#ifdef DEBUG
	sprintf( errmsg, "encode_hexstr: found '%s' on line %d\n", helper->strval, line );
	errmsg += strlen(errmsg);
#endif /* DEBUG */ 
	return( i );  
}

/* This is for strings which need the terminating 0 at the end, e.g. Service Flow Class Name */

int encode_oid(unsigned char *buf, void *tval, symbol_type *sym_ptr ) {
	unsigned int	output_size;
	union t_val	*helper;

	if ( buf == NULL ) {
		sprintf( errmsg, "encode_oid called w/NULL buffer!\n");
		errmsg += strlen(errmsg);
		return 0;
	}

	if ( tval == NULL  ) {
		sprintf( errmsg, "encode_oid called w/NULL value struct !\n");
		errmsg += strlen(errmsg);
		return 0;
	}

	helper = (union t_val *) tval;

	output_size = encode_snmp_oid( helper->strval, buf, TLV_VSIZE ); 
	return( output_size );
}


int encode_nothing(unsigned char *buf, void *tval, symbol_type *sym_ptr ) {
	return 0;
}


/***********************************************************************************/
/***********************************************************************************/
/***********************************************************************************/


/* 
 * Given a symbol identifier, AND a buffer containing the raw "value", 
 * returns a pointer to a tlv structure containing TLV buffer and the 
 * length of the buffer, including docsis code and value length fields. 
 * The length is relevant when we "assemble" two tlv_lists from an  
 * an assignment_list + assignment_stmt
 */ 

struct tlv *create_tlv( symbol_type *sym_ptr, union t_val *value ) {
	struct tlv	*tlvbuf = NULL;

	tlvbuf = (struct tlv *) calloc(1, sizeof(struct tlv) ) ;	
	tlvbuf->docs_code = sym_ptr->docsis_code;
	tlvbuf->tlv_len = sym_ptr->encode_func(tlvbuf->tlv_value,value,sym_ptr);
	if (tlvbuf->tlv_len <= 0 ) { 
		sprintf( errmsg, "Got 0-length value while scanning for %s at line %d\n",
				sym_ptr->sym_ident, line );
		errmsg += strlen(errmsg);
		tlvbuf->docs_code = 0;
		tlvbuf->tlv_len = 0;
	}
	return tlvbuf;
}

/* Given a symbol identifier ( well, always SnmpMibObject ), a string 
** containing an OID name, a string which sets the value of the OID and a t_val 
** which contains the raw value (of type oid_asntype), creates a TLV with the 
** corresponding VariableBinding  
*/

struct tlv *create_snmpset_tlv ( symbol_type *sym_ptr, char *oid_string,
				char oid_asntype, union t_val *value ) {
	struct tlv	*tlvbuf = NULL;

	tlvbuf = (struct tlv *) calloc(1, sizeof(struct tlv));	
	tlvbuf->docs_code = sym_ptr->docsis_code;
	tlvbuf->tlv_len = encode_vbind (oid_string, oid_asntype, value,
					tlvbuf->tlv_value, TLV_VSIZE );

	if (tlvbuf->tlv_len <= 0 ) { 
		sprintf( errmsg, "got len 0 value while scanning for %s\n at line %d",
			sym_ptr->sym_ident, line );
		errmsg += strlen(errmsg);
		tlvbuf->docs_code = 0;
		tlvbuf->tlv_len = 0;
	}
	return tlvbuf;
}


/* Given a symbol identifier ( always SnmpWrite ), a string
** containing an OID, and an integer, creates a TLV containing the  
** OID ASN encoding + 1 control byte to permit/deny SNMP Write operations on a
** MIB variable which has the OID as prefix.
*/

struct tlv *create_snmpw_tlv ( symbol_type *sym_ptr, char *oid_string,
				union t_val *value ) {
	struct tlv	*tlvbuf=NULL;

	tlvbuf = (struct tlv *) calloc(1, sizeof(struct tlv));
	tlvbuf->docs_code = sym_ptr->docsis_code;
	tlvbuf->tlv_len = encode_snmp_oid ( oid_string, tlvbuf->tlv_value, TLV_VSIZE );

	if (tlvbuf->tlv_len <= 0 ) {
		sprintf(errmsg, "got len 0 value while scanning for %s\n at line %d",
			sym_ptr->sym_ident, line );
		errmsg += strlen(errmsg);
		tlvbuf->docs_code = 0;
		tlvbuf->tlv_len = 0;
	} else {
		tlvbuf->tlv_value[tlvbuf->tlv_len] = (unsigned char) value->intval ;
		tlvbuf->tlv_len++;
	}
	return tlvbuf;
}         

/* Given a code, length and raw value (encoded as a hex string), 
** creates a TLV encoding. This can be used for e.g. VendorSpecific 
** Information or unsupported (future?) configuration settings. 
*/
struct tlv *create_generic_tlv ( symbol_type *sym_ptr, int tlv_code,
				int tlv_length, union t_val *value ) {
	struct tlv	*tlvbuf = NULL;

	tlvbuf = (struct tlv *) calloc(1, sizeof(struct tlv));
	tlvbuf->docs_code = tlv_code;
	tlvbuf->tlv_len = encode_hexstr ( tlvbuf->tlv_value, value, sym_ptr );

	if (tlvbuf->tlv_len <= 0 ) {
		sprintf( errmsg, "got len 0 value while scanning for %s\n at line %d",
			sym_ptr->sym_ident, line );
		errmsg += strlen(errmsg);
		tlvbuf->docs_code = 0;
		tlvbuf->tlv_len = 0;
	}
	if (tlvbuf->tlv_len != tlv_length ) {
		sprintf( errmsg, "Length mismatch while encoding GenericTLV: ");
		errmsg += strlen(errmsg);
		sprintf( errmsg, "given length %d, value value length %d at line %d",
				tlvbuf->tlv_len, tlv_length, line );
		errmsg += strlen(errmsg);
		tlvbuf->docs_code = 0;
		tlvbuf->tlv_len = 0;
	}
	return tlvbuf;
}

/* Adds a TLV to a tlvlist. If the tlvlist pointer we are called with is NULL, 
 * we create the new tlvlist. 
 */ 

struct tlv_list *add_tlv_to_list(struct tlv_list *list, struct tlv *newtlv) {
	struct tlv_list *newtlvlist;

	newtlvlist = (struct tlv_list *) calloc(1,  sizeof(struct tlv_list) );
	if ( list == NULL ) {  /* assignment_list :== assignment_stmt  */
		/* create list of tlv pointers with 1 tlv pointer */
		newtlvlist->tlv_count = 1;	
		newtlvlist->tlvlist   = (struct tlv **) calloc(1,  sizeof (struct tlv *));
  		newtlvlist->tlvlist[0] = newtlv;
		return newtlvlist;
	} else { /* assignment_list :== assignment_list assignment_stmt */
		/* add a new tlv pointer to the existing list of tlv pointers */
		newtlvlist->tlv_count = list->tlv_count + 1;
		newtlvlist->tlvlist = (struct tlv **) calloc(1, (newtlvlist->tlv_count)*sizeof(struct tlv *));
		memcpy(&(newtlvlist->tlvlist[0]),&(list->tlvlist[0]),list->tlv_count*sizeof(struct tlv *));
		newtlvlist->tlvlist[newtlvlist->tlv_count-1]=newtlv;
		free (list->tlvlist); free(list);
		return newtlvlist;
	}
}

/* Merges two tlvlists into one tlvlist. */

struct tlv_list *merge_tlvlist( struct tlv_list *list1, struct tlv_list *list2 ) {
	struct tlv_list *newtlvlist;

	if ( list1 != NULL && list2 == NULL ) return list1;
	if ( list1 == NULL && list2 != NULL ) return list2;

	if ( list1 == NULL && list2 == NULL ) {
		sprintf( errmsg, "merge_tlvlist called with 2 NULL lists !\n"); 
		errmsg += strlen(errmsg);
		return NULL;
	}
  
	newtlvlist = (struct tlv_list *) calloc(1,  sizeof(struct tlv_list) );
	newtlvlist->tlv_count = list1->tlv_count + list2->tlv_count;
	newtlvlist->tlvlist = (struct tlv **) calloc(1, (newtlvlist->tlv_count)*sizeof ( struct tlv *));
	memcpy ( &(newtlvlist->tlvlist[0]), &(list1->tlvlist[0]),
		list1->tlv_count * sizeof(struct tlv *) );
	memcpy ( &(newtlvlist->tlvlist[list1->tlv_count]), &(list2->tlvlist[0]),
		list2->tlv_count*sizeof(struct tlv *));
	free(list1->tlvlist); free(list1);
	free(list2->tlvlist); free(list2);
	/* We don't free the TLVs since they are still referenced by newtlvlist->tlvlist */
	return newtlvlist;
}

/* 
 * Used for creating a top-level list corresponding to ClassOfService,
 * BaselinePrivacy, Telephony Settings etc.
 * Creates a "flat" TLV buffer out of a tlvlist, and assembles the 
 * TLV buffer into a new TLV which has the parent's docsis_code. 
 * Returns a TLV list with the new TLV 
 */

struct tlv_list *assemble_list_in_parent (symbol_type *sym_ptr, struct tlv_list *list) {
	struct tlv_list	*newtlvlist;
	struct tlv	*parent_tlv;

	newtlvlist = (struct tlv_list *) calloc(1,  sizeof ( struct tlv_list ));
	newtlvlist->tlvlist = (struct tlv **) calloc(1,  sizeof ( struct tlv *));
	parent_tlv = (struct tlv *) calloc(1,  sizeof ( struct tlv) );
  
	parent_tlv->docs_code = sym_ptr->docsis_code;
	parent_tlv->tlv_len = flatten_tlvlist(parent_tlv->tlv_value,list);
	newtlvlist->tlv_count = 1;
	newtlvlist->tlvlist[0] = parent_tlv;
	return newtlvlist;
}

/* Creates a buffer filled with the TLV bytes sequentially, as they will be 
 * found in the final configuration file. 
 * We use this function either to create a "parent" tlv corresponding to 
 * ClassOfService-like TLVs, or to create the final buffer to which we add the 
 * MIC digests, end-of-data marker and pad and write to the output file.
 */

unsigned int flatten_tlvlist (unsigned char *buf, struct tlv_list *list ) {
	unsigned char	*cp;
	int		i;

	if ( buf == NULL ) { 
		sprintf(errmsg, "Error: can't flatten tlvlist in a NULL destination  buffer!\n" );
		errmsg += strlen(errmsg);
		return 0;
	}
	cp = buf;

	for ( i=0; i < list->tlv_count; i++ ) { 
		*cp = (unsigned char) list->tlvlist[i]->docs_code;  cp++;
		*cp = (unsigned char) list->tlvlist[i]->tlv_len;    cp++;
		memcpy ( cp, list->tlvlist[i]->tlv_value, list->tlvlist[i]->tlv_len ); 
		cp = cp + list->tlvlist[i]->tlv_len;
		free ( list->tlvlist[i] ); /* The TLV won't be referenced anymore */
	}
	/* We leave it to the caller to free the list */
	return (unsigned int) (cp-buf); 
}


