
/* 
 *  DOCSIS configuration file encoder. 
 *  Copyright (c) 2001 Cornel Ciocirlan, ctrl@users.sourceforge.net.
 *  Copyright (c) 2002 Evvolve Media SRL,office@evvolve.com
 *  
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *  DOCSIS is a registered trademark of Cablelabs, http://www.cablelabs.com
 */

#ifndef _DOCSIS_ENCODE_H
#define _DOCSIS_ENCODE_H

#define TLV_VSIZE 255

struct tlv {
        unsigned char docs_code;
        unsigned char tlv_len;
        unsigned char tlv_value[TLV_VSIZE];
};

struct tlv_list {
        int tlv_count;  /* How many TLVs we have in this list */
        struct tlv **tlvlist;
};

union t_val {           /* union for returning token values */
        int intval;             /* For integers */
        unsigned int uintval;             /* For integers */
        symbol_type *symptr;    /* For token identifiers */
        char *strval;           /* For strings */
        unsigned char *ustrval; /* For strings */
        unsigned int ip;        /* For IP Addresses */
        struct tlv_list *tlvlist; /* For for struct tlvlist pointers */
        struct tlv *tlvptr;     /* For struct tlv pointers; */
};


/* each of these returns the length of the value parsed ! */

int encode_uint   	(unsigned char *buf, void *tval, symbol_type *sym_ptr ); 
int encode_ushort 	(unsigned char *buf, void *tval, symbol_type *sym_ptr ); 
int encode_uchar  	(unsigned char *buf, void *tval, symbol_type *sym_ptr );
int encode_ip     	(unsigned char *buf, void *tval, symbol_type *sym_ptr );
int encode_ether	(unsigned char *buf, void *tval, symbol_type *sym_ptr );
int encode_ethermask	(unsigned char *buf, void *tval, symbol_type *sym_ptr );
int encode_string 	(unsigned char *buf, void *tval, symbol_type *sym_ptr );
int encode_strzero	(unsigned char *buf, void *tval, symbol_type *sym_ptr );
int encode_hexstr 	(unsigned char *buf, void *tval, symbol_type *sym_ptr );
int encode_oid 		(unsigned char *buf, void *tval, symbol_type *sym_ptr );
int encode_nothing	(unsigned char *buf, void *tval, symbol_type *sym_ptr );


struct tlv *create_tlv(symbol_type *sym_ptr, union t_val *value);

struct tlv *create_snmpset_tlv ( symbol_type *sym_ptr, char *oid_string,
				char oid_asntype, union t_val *value );

struct tlv *create_snmpw_tlv ( symbol_type *sym_ptr, char *oid_string,
                               union t_val *value );

struct tlv *create_generic_tlv ( symbol_type *sym_ptr, int tlv_code,
				int tlv_length, union t_val *value );

struct tlv_list *add_tlv_to_list(struct tlv_list *list, struct tlv *newtlv);

struct tlv_list *merge_tlvlist( struct tlv_list *list1, struct tlv_list *list2 );

struct tlv_list *assemble_list_in_parent (symbol_type *sym_ptr, struct tlv_list *list);

unsigned int flatten_tlvlist (unsigned char *buf, struct tlv_list *list );



#endif /* _DOCSIS_ENCODE_H */

