/* 
 *  DOCSIS configuration file encoder. 
 *  Copyright (c) 2001 Cornel Ciocirlan, ctrl@users.sourceforge.net.
 *  Copyright (c) 2002 Evvolve Media SRL,office@evvolve.com
 *  
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *  DOCSIS is a registered trademark of Cablelabs, http://www.cablelabs.com
 */

/*\.([0-9]+\.)*[0-9]+		{ TSAVE(yytext);yylval.strval=tsave; return T_OID; 	  } */

%{ 
#include "docsis_symtable.h"
#include "docsis_decode.h"
#include "docsis_encode.h"
#include "docsis_yy.tab.h"
#include "docsis_snmp.h"

#ifdef DEBUG
extern int yydebug;
#endif

extern char *errmsg;

unsigned int line=1; /* On parse errors, we print the line number */

char *str; char strbuf[255]; unsigned int strsize;/* max string len = 255; TODO check overflow */

char *tsave;
#define TSAVE2  {					\
	yylval.strval = (char*) calloc(yyleng+1,1);	\
	memcpy( yylval.strval, yytext, yyleng );	\
	}						\

#define TSAVE(s) tsave=(char*)calloc(1,yyleng+1);memcpy(tsave,s,yyleng)
%}
%x STRING
%x COMMENT
%%

([0-9]+\.){3}[0-9]+	 	{ TSAVE2; return T_IP;	  }
0[Xx]([0-9A-Fa-f])+		{ TSAVE2; return T_HEX_STRING; }
([0-9A-Fa-f][0-9A-Fa-f]:){5}[0-9A-Fa-f][0-9A-Fa-f]\/([0-9A-Fa-f][0-9A-Fa-f]:){5}[0-9A-Fa-f][0-9A-Fa-f]  { TSAVE2; return T_ETHERMASK; 			       }
([0-9A-Fa-f][0-9A-Fa-f]:){5}[0-9A-Fa-f][0-9A-Fa-f]  { TSAVE2; return T_MAC;}

(\.{1})*([A-Za-z0-9_-]+\.)+[A-Za-z0-9]+	{ TSAVE2; return T_LABEL_OID; }
(\.{1})*([A-Za-z0-9_-]+\.)+'[A-Za-z0-9,:_\-\.]+'	{ TSAVE2; return T_LABEL_OID; }
(\.{1})*([A-Za-z0-9_-]+\.)+((\"{1})*[A-Za-z0-9,:_\-\.]+(\"{1})*)+	{ TSAVE2; return T_LABEL_OID; }
[Mm][Aa][Ii][Nn]	{ TSAVE2; return T_MAIN; }
[Ss][Nn][Mm][Pp][Ww][Rr][Ii][Tt][Ee][Cc][Oo][Nn][Tt][Rr][Oo][Ll] {
	yylval.symptr = find_symbol_by_name(yytext);
				if (yylval.symptr == NULL) {
					sprintf(errmsg,"line %d: bad SnmpWriteControl \n",line);
					errmsg += strlen(errmsg);
					break;
				}
				return T_IDENT_SNMPW;
			}
[Ss][Nn][Mm][Pp][Mm][Ii][Bb][Oo][Bb][Jj][Ee][Cc][Tt]	{
	yylval.symptr = find_symbol_by_name(yytext);
				if (yylval.symptr == NULL) {
					sprintf(errmsg,"line %d: bad SnmpMibObject \n",line);
					errmsg += strlen(errmsg);
					break;
				}
				return T_IDENT_SNMPSET;
			}
[Ii][Nn][Tt][Ee][Gg][Ee][Rr]		{
	return T_ASNTYPE_INT; 	}
[Uu][Nn][Ss][Ii][Gg][Nn][Ee][Dd][Ii][Nn][Tt][Ee][Gg][Ee][Rr]	{
	return T_ASNTYPE_UINT; 	}
[Gg][Aa][Uu][Gg][Ee]			{
	return T_ASNTYPE_GAUGE; 	}
[Tt][Ii][Mm][Ee][Tt][Ii][Cc][Kk][Ss]	{
	return T_ASNTYPE_TICKS; 	}
[Ii][Pp][Aa][Dd][Dd][Rr][Ee][Ss][Ss]	{
	return T_ASNTYPE_IP; 		}
[Oo][Bb][Jj][Ee][Cc][Tt][Ee][Dd]	{
	return T_ASNTYPE_OBJID; 	}
[Ss][Tt][Rr][Ii][Nn][Gg]		{
	return T_ASNTYPE_STRING; 	}
[Hh][Ee][Xx][Ss][Tt][Rr][Ii][Nn][Gg]	{
	return T_ASNTYPE_HEXSTR; 	}
[Dd][Ee][Cc][Ii][Mm][Aa][Ll][Ss][Tt][Rr][Ii][Nn][Gg]	{
	return T_ASNTYPE_DECSTR; 	}
[Bb][Ii][Tt][Ss][Tt][Rr][Ii][Nn][Gg]	{
	return T_ASNTYPE_BITSTR; 	}
[Bb][Ii][Gg][Ii][Nn][Tt]		{
	return T_ASNTYPE_BIGINT; 	}
[Uu][Nn][Ss][Ii][Gg][Nn][Ee][Dd][Bb][Ii][Gg][Ii][Nn][Tt]	{
	return T_ASNTYPE_UBIGINT; 	}
[Ff][Ll][Oo][Aa][Tt]			{
	return T_ASNTYPE_FLOAT; 	}
[Dd][Oo][Uu][Bb][Ll][Ee]		{
	return T_ASNTYPE_DOUBLE; 	}
[Gg][Ee][Nn][Ee][Rr][Ii][Cc][Tt][Ll][Vv]	{
	yylval.symptr = find_symbol_by_name(yytext);
				if (yylval.symptr == NULL) {
					sprintf(errmsg,"line %d: bad GenericTLV \n",line);
					errmsg += strlen(errmsg);
					break;
				}
				return T_IDENT_GENERIC;
			}
[Tt][Ll][Vv][Cc][Oo][Dd][Ee]	{
	return T_TLV_CODE; 	}
[Tt][Ll][Vv][Ll][Ee][Nn][Gg][Tt][Hh]	{
	return T_TLV_LENGTH; 	}
[Tt][Ll][Vv][Vv][Aa][Ll][Uu][Ee]	{
	return T_TLV_VALUE; 	}
[A-Za-z]+[A-Za-z0-9]*	{ 
				yylval.symptr = find_symbol_by_name(yytext);
				if (yylval.symptr == NULL) {
					sprintf(errmsg,"line %d: unknown command \n",line);
					errmsg += strlen(errmsg);
					break;
				}
#ifdef DEBUG
		  	  	if(yydebug) { sprintf(errmsg, "yylex:returning T_IDENTIFIER, yylval.symptr->sym_ident=%s\n", 
							yylval.symptr->sym_ident);
					errmsg += strlen(errmsg); }
#endif
			 	return T_IDENTIFIER; 
			}

\"		{ BEGIN STRING; str=strbuf; memset(strbuf,0,255); strsize=0; }	
<STRING>\\ 	{ 
			if (strsize<255)  
			{
				*str++ = ' '; 
				strsize++;
			} else {
			  sprintf(errmsg,"line %d: string too long (max 255 characters)\n",line);
			  errmsg += strlen(errmsg);
			}
		}
<STRING>\;	{
			sprintf(errmsg,"line %d: ';' not allowed in string, unmatched \" ?\n",line);
			errmsg += strlen(errmsg);
		}
<STRING>\n	{
			sprintf(errmsg,"line %d: \\n not allowed in string, unmatched \" ?\n",line);
			errmsg += strlen(errmsg);
		}
<STRING>\t	{
			sprintf(errmsg,"line %d: \\t not allowed in string, unmatched \" ?\n",line);
			errmsg += strlen(errmsg);
		}
<STRING>\"	{ *str=0; BEGIN 0; yylval.strval=strbuf; return T_STRING; }
<STRING>.	{ 
			if (strsize<255)  
			{
				*str++ = *yytext; 
				strsize++;
			} else {
			  sprintf(errmsg,"line %d: string too long (max 255 characters)\n",line);
			  errmsg += strlen(errmsg);
			}
		}
\/[*]+		{ BEGIN COMMENT; }
<COMMENT>\n	{ line ++; }
<COMMENT>.	
<COMMENT>\/[*]+	{ sprintf (errmsg,"line %d: comment-in-comment not supported", line); errmsg += strlen(errmsg);}
<COMMENT>[*]+\/	{ BEGIN 0; }

[0-9]+	 { yylval.intval = atoi(yytext); return T_INTEGER; }
-[0-9]+	 { yylval.intval = atoi(yytext); return T_INTEGER; }
[ \t]+	 ;
\r\n	 { line++; }
\n	 { line++; }
\{ 	 { TSAVE2; return '{'; }
\}	 { TSAVE2; return '}'; }
;	 { TSAVE2; return ';'; }
.	 {
		sprintf(errmsg,"Unrecognized char \"%c\" at line %d\n",*yytext,line);
		errmsg += strlen(errmsg);
	}
%%
