/* 
 *  DOCSIS configuration file encoder. 
 *  Copyright (c) 2001 Cornel Ciocirlan, ctrl@users.sourceforge.net.
 *  Copyright (c) 2002,2003 Evvolve Media SRL,office@evvolve.com
 *  
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *  DOCSIS is a registered trademark of Cablelabs, http://www.cablelabs.com
 */

#include <stdio.h>
#include <string.h>
#include <sys/types.h>

#include "docsis_symtable.h"
#include "docsis_encode.h"
#include "docsis_yy.tab.h"
#include "docsis_snmp.h"


#include <net-snmp/net-snmp-config.h>
#include <net-snmp/utilities.h>
#include <net-snmp/config_api.h>
#include <net-snmp/output_api.h>
#include <net-snmp/mib_api.h>

extern unsigned int line;	/* from a.l */
extern char *errmsg;
extern int omx;		/* external output buffer max */

#define PACKET_LENGTH (8 * 1024)



/* Prototypes for Utility Function */
int hexadecimal_to_binary( const char *str, unsigned char * bufp );



unsigned int  encode_vbind( char *oid_string, char oid_asntype, union t_val *value,
			    unsigned char *out_buffer, unsigned int out_size )
{
  oid var_name[MAX_OID_LEN];
  size_t name_len = MAX_OID_LEN;
  unsigned char *data_ptr;
  unsigned char buf[SPRINT_MAX_LEN];
  unsigned int ltmp;
  struct tree *tp;
  struct range_list *rp;
  long len=0;
  int rv;

	// for debuging
	// fprintf(stderr, "now encoding %s\n", oid_string );

  memset (buf, 0, SPRINT_MAX_LEN);
  if (!get_node (oid_string, var_name, &name_len))
    {
      if (!read_objid (oid_string, var_name, &name_len))
	{
	  /* enable random access */
	  if (!netsnmp_ds_get_boolean
	      (NETSNMP_DS_LIBRARY_ID, NETSNMP_DS_LIB_RANDOM_ACCESS))
	    {
	      netsnmp_ds_toggle_boolean (NETSNMP_DS_LIBRARY_ID,
					 NETSNMP_DS_LIB_RANDOM_ACCESS);
	    }

	  if (!get_node (oid_string, var_name, &name_len))
	    {
	      /* Not found, try UCD style OID access */
	      netsnmp_ds_toggle_boolean (NETSNMP_DS_LIBRARY_ID,
					 NETSNMP_DS_LIB_READ_UCD_STYLE_OID);
	      if (!netsnmp_ds_get_boolean
		  (NETSNMP_DS_LIBRARY_ID, NETSNMP_DS_LIB_READ_UCD_STYLE_OID))
		{
		  netsnmp_ds_toggle_boolean (NETSNMP_DS_LIBRARY_ID,
					     NETSNMP_DS_LIB_READ_UCD_STYLE_OID);
		}

	      if (!read_objid (oid_string, var_name, &name_len))
		{

		  /* Hmm, still not found, go back to net-snmp-style OID and try regex access */
		  netsnmp_ds_toggle_boolean (NETSNMP_DS_LIBRARY_ID,
					     NETSNMP_DS_LIB_READ_UCD_STYLE_OID);

		  if (!netsnmp_ds_get_boolean
		      (NETSNMP_DS_LIBRARY_ID, NETSNMP_DS_LIB_REGEX_ACCESS))
		    {
		      netsnmp_ds_toggle_boolean (NETSNMP_DS_LIBRARY_ID,
						 NETSNMP_DS_LIB_REGEX_ACCESS);
		    }

		  if (!get_node (oid_string, var_name, &name_len))
		    {
		      /* Ran out of ideas. */
		      sprintf(errmsg, "Error: Can't find oid %s at line %d \n",
			      oid_string, line);
			errmsg += strlen( errmsg );
			return 0;
		    }
		}
	    }
	}
    }

  switch (oid_asntype)
    {
    case 'i':
      data_ptr = snmp_build_var_op (out_buffer,
				    var_name,
				    &name_len,
				    ASN_INTEGER,
				    sizeof (int),
				    (unsigned char *) &value->intval,
				    &out_size);
      return data_ptr - out_buffer;
      break;
      ;;
    case 'g':
	/* TODO: check that the int is positive ? */

      data_ptr = snmp_build_var_op (out_buffer,
				    var_name,
				    &name_len,
				    ASN_GAUGE,
				    sizeof (int),
				    (unsigned char *) &value->intval,
				    &out_size);
      return data_ptr - out_buffer;
      break;
      ;;
    case 's':
    case 'x':
    case 'd':
      if (oid_asntype == 'd')
	{
	  /* ltmp = ascii_to_binary(value->strval, buf);
	     strncpy((char*)buf, value, SPRINT_MAX_BUF);
	     ltmp = strlen((char*)buf); */
	}
      else if (oid_asntype == 's')
	{
	  strncpy ((char *) buf, value->strval, SPRINT_MAX_LEN);
	  len = strlen ((char *) buf);
	}
      else if (oid_asntype == 'x')
	{
	  if ((rv = hexadecimal_to_binary (value->strval, buf)) == -1)
	    {
	      sprintf(errmsg, "Invalid hexadecimal string at line %d\n", line);
		errmsg += strlen( errmsg );
		return 0;
	    }
	  ltmp = (unsigned int) rv;
	  len = ltmp;
	}

      if (len < 0 || len > SPRINT_MAX_LEN - 1)
	{
	  sprintf(errmsg, "String too long at line %d, max allowed %d\n", line,
		  SPRINT_MAX_LEN);
	  errmsg += strlen( errmsg );
	  return 0;
	  break;
	}
      tp = get_tree (var_name, name_len, get_tree_head ());
      if (tp && tp->ranges
	  && !netsnmp_ds_get_boolean (NETSNMP_DS_LIBRARY_ID,
				      NETSNMP_DS_LIB_DONT_CHECK_RANGE))
	{
	  rp = tp->ranges;
	  while (rp)
	    {
	      if (rp->low <= len && len <= rp->high)
		break;
	      rp = rp->next;
	    }
	  if (!rp)
	    {
	      sprintf(errmsg, "Value too long at line %d\n", line);
		errmsg += strlen( errmsg );
		return 0;
	      break;
	    }
	}
      data_ptr = snmp_build_var_op (out_buffer,
				    var_name,
				    &name_len,
				    ASN_OCTET_STR,
				    len, (unsigned char *) buf, &out_size);
      return data_ptr - out_buffer;
      break;
    case 'a':
      if (!inet_aton (value->strval, (struct in_addr *) &ltmp))
	{
	  sprintf(errmsg, "Invalid IP address %s at line %d\n", value->strval, line);
	  errmsg += strlen( errmsg );
	  return 0;
	}
      data_ptr = snmp_build_var_op (out_buffer,
				    var_name,
				    &name_len,
				    ASN_IPADDRESS,
				    sizeof (int),
				    (unsigned char *) &ltmp, &out_size);
      return data_ptr - out_buffer;
      break;
      ;;
    default:
      sprintf(errmsg, "Variable type %s not supported yet", &oid_asntype);
	errmsg += strlen( errmsg );
      return 0;
      ;;
    }
/* NOTREACHED */
}


int decode_vbind (unsigned char *data, unsigned int vb_len, char *out_buffer ) {
	unsigned char	*var_val;
	oid		var_name[MAX_OID_LEN];	/* To test the objid */
	size_t		name_len = MAX_OID_LEN;	/* To test the objid */
	int		badtype = 0;
	unsigned int	len;
	struct variable_list	*vp;
	oid		objid[MAX_OID_LEN];
	char		_docsis_snmp_label[50];	/* To hold the 'name' of the type, i.e. Integer etc */
	char		*enum_string = NULL;
	struct tree	*subtree;
	struct		enum_list *enums;

	vp = (struct variable_list *) calloc(1, sizeof (struct variable_list));
	if (vp == NULL) { return 0; }

	// vp->next_variable = NULL;
	// vp->val.string = NULL;
	vp->name_length = MAX_OID_LEN;
	vp->name = 0;
	data = snmp_parse_var_op( data, objid, &vp->name_length, &vp->type,
				  &vp->val_len, &var_val, (size_t *) & vb_len);

	if (data == NULL) return -1;
	if ( snmp_set_var_objid( vp, objid, vp->name_length) ) return -1;

	len = PACKET_LENGTH;

	if( netsnmp_ds_get_boolean(NETSNMP_DS_LIBRARY_ID, NETSNMP_DS_LIB_EXTENDED_INDEX) ) {
		netsnmp_ds_toggle_boolean (NETSNMP_DS_LIBRARY_ID, NETSNMP_DS_LIB_EXTENDED_INDEX);
	} /* Disable extended index format ... makes it harder to parse tokens in lex */

	if (!netsnmp_ds_get_boolean(NETSNMP_DS_LIBRARY_ID, NETSNMP_DS_LIB_PRINT_NUMERIC_ENUM)) {
		netsnmp_ds_toggle_boolean (NETSNMP_DS_LIBRARY_ID,NETSNMP_DS_LIB_PRINT_NUMERIC_ENUM);
	} /* Enable printing numeric enums */

	if (netsnmp_ds_get_boolean(NETSNMP_DS_LIBRARY_ID, NETSNMP_DS_LIB_ESCAPE_QUOTES)) {
		netsnmp_ds_toggle_boolean (NETSNMP_DS_LIBRARY_ID,NETSNMP_DS_LIB_ESCAPE_QUOTES);
	} /* Disable escape quotes in string index output  */
 
	snprint_objid( out_buffer, omx, vp->name, vp->name_length);

	if ( !get_node(out_buffer, var_name, &name_len)) {
		if (!read_objid (out_buffer, var_name, &name_len)) {
			sprintf(errmsg, "/* Hmm ... can't find oid %s at line %d ", out_buffer, line);
			errmsg += strlen( errmsg );
			sprintf(errmsg, "... perhaps the MIBs are not installed ? */\n" );
			errmsg += strlen( errmsg );

			/* temporarily set full output format */
			netsnmp_ds_set_int (NETSNMP_DS_LIBRARY_ID,
					    NETSNMP_DS_LIB_OID_OUTPUT_FORMAT,
					    NETSNMP_OID_OUTPUT_FULL);
			memset (out_buffer, 0, omx);
			snprint_objid (out_buffer, omx, vp->name, vp->name_length);

			/* Go back to suffix-mode for better readability */
			netsnmp_ds_set_int (NETSNMP_DS_LIBRARY_ID,
					    NETSNMP_DS_LIB_OID_OUTPUT_FORMAT,
					    NETSNMP_OID_OUTPUT_SUFFIX);
		}
	}

	/* Flush Buffer */
	omx -= strlen( out_buffer );
	out_buffer += strlen( out_buffer );
	sprintf(out_buffer,"\t");
	omx--; out_buffer++;

	/* save the subtree - we need it later to show enums */
	subtree = get_tree (var_name, name_len, get_tree_head() );

	/* This first switch is just for saving the type in the
	   format we actually want to print. */

  switch ((short) vp->type)
    {
    case ASN_INTEGER:
      memset (_docsis_snmp_label, 0, 50);
      sprintf (_docsis_snmp_label, "Integer");
      break;
    case ASN_COUNTER:
      memset (_docsis_snmp_label, 0, 50);
      sprintf (_docsis_snmp_label, "Counter");
      break;
    case ASN_GAUGE:
      memset (_docsis_snmp_label, 0, 50);
      sprintf (_docsis_snmp_label, "Gauge");
      break;
    case ASN_TIMETICKS:
      memset (_docsis_snmp_label, 0, 50);
      sprintf (_docsis_snmp_label, "TimeTicks");
      break;
    case ASN_UINTEGER:
      memset (_docsis_snmp_label, 0, 50);
      sprintf (_docsis_snmp_label, "UnsignedInteger");
      break;
#ifdef OPAQUE_SPECIAL_TYPES
    case ASN_OPAQUE_COUNTER64:
      memset (_docsis_snmp_label, 0, 50);
      sprintf (_docsis_snmp_label, "OpaqueCounter64");
      break;
    case ASN_OPAQUE_U64:
      memset (_docsis_snmp_label, 0, 50);
      sprintf (_docsis_snmp_label, "OpaqueU64");
      break;
#endif /* OPAQUE_SPECIAL_TYPES */
    case ASN_COUNTER64:
      memset (_docsis_snmp_label, 0, 50);
      sprintf (_docsis_snmp_label, "Counter64");
      break;
#ifdef OPAQUE_SPECIAL_TYPES
    case ASN_OPAQUE_FLOAT:
      memset (_docsis_snmp_label, 0, 50);
      sprintf (_docsis_snmp_label, "OpaqueFloat");
      break;
    case ASN_OPAQUE_DOUBLE:
      memset (_docsis_snmp_label, 0, 50);
      sprintf (_docsis_snmp_label, "OpaqueDouble");
      break;
    case ASN_OPAQUE_I64:
      memset (_docsis_snmp_label, 0, 50);
      sprintf (_docsis_snmp_label, "OpaqueI64");
      break;
#endif /* OPAQUE_SPECIAL_TYPES */
    case ASN_OCTET_STR:
      memset (_docsis_snmp_label, 0, 50);
      sprintf (_docsis_snmp_label, "String");
      break;
    case ASN_IPADDRESS:
      memset (_docsis_snmp_label, 0, 50);
      sprintf (_docsis_snmp_label, "IPAddress");
      break;
    case ASN_OPAQUE:
      memset (_docsis_snmp_label, 0, 50);
      sprintf (_docsis_snmp_label, "Opaque");
      break;
    case ASN_NSAP:
      memset (_docsis_snmp_label, 0, 50);
      sprintf (_docsis_snmp_label, "NSAP");
      break;
    case ASN_OBJECT_ID:
      memset (_docsis_snmp_label, 0, 50);
      sprintf (_docsis_snmp_label, "ObjectID");
      break;
    case ASN_BIT_STR:
      memset (_docsis_snmp_label, 0, 50);
      sprintf (_docsis_snmp_label, "BitString");
      break;
    }

	snprintf(out_buffer,omx," %s ", _docsis_snmp_label );
	omx -= strlen( out_buffer );
	out_buffer += strlen( out_buffer );

  switch ((short) vp->type)
    {
    case ASN_INTEGER:

      vp->val.integer = (long *) vp->buf;
      vp->val_len = sizeof (long);
      asn_parse_int (var_val, &len, &vp->type,
		     (long *) vp->val.integer, sizeof (vp->val.integer));

      break;
    case ASN_COUNTER:
    case ASN_GAUGE:
    case ASN_TIMETICKS:
    case ASN_UINTEGER:
      vp->val.integer = (long *) vp->buf;
      vp->val_len = sizeof (u_long);
      asn_parse_unsigned_int (var_val, &len, &vp->type,
			      (u_long *) vp->val.integer,
			      sizeof (vp->val.integer));

      break;

#ifdef OPAQUE_SPECIAL_TYPES
    case ASN_OPAQUE_COUNTER64:
    case ASN_OPAQUE_U64:
#endif /* OPAQUE_SPECIAL_TYPES */
    case ASN_COUNTER64:
      vp->val.counter64 = (struct counter64 *) vp->buf;
      vp->val_len = sizeof (struct counter64);
      asn_parse_unsigned_int64 (var_val, &len, &vp->type,
				(struct counter64 *) vp->val.counter64,
				sizeof (*vp->val.counter64));
      break;
#ifdef OPAQUE_SPECIAL_TYPES
    case ASN_OPAQUE_FLOAT:
      vp->val.floatVal = (float *) vp->buf;
      vp->val_len = sizeof (float);
      asn_parse_float (var_val, &len, &vp->type,
		       vp->val.floatVal, vp->val_len);
      break;
    case ASN_OPAQUE_DOUBLE:
      vp->val.doubleVal = (double *) vp->buf;
      vp->val_len = sizeof (double);
      asn_parse_double (var_val, &len, &vp->type,
			vp->val.doubleVal, vp->val_len);
      break;
    case ASN_OPAQUE_I64:
      vp->val.counter64 = (struct counter64 *) vp->buf;
      vp->val_len = sizeof (struct counter64);
      asn_parse_signed_int64 (var_val, &len, &vp->type,
			      (struct counter64 *) vp->val.counter64,
			      sizeof (*vp->val.counter64));

      break;
#endif /* OPAQUE_SPECIAL_TYPES */
    case ASN_OCTET_STR:
    case ASN_IPADDRESS:
    case ASN_OPAQUE:
    case ASN_NSAP:
      if (vp->val_len < sizeof (vp->buf))
	{
	  vp->val.string = (u_char *) vp->buf;
	}
      else
	{
	  vp->val.string = (u_char *) calloc(1, (unsigned) vp->val_len);
	}
      asn_parse_string (var_val, &len, &vp->type, vp->val.string,
			&vp->val_len);
      break;
    case ASN_OBJECT_ID:
      vp->val_len = MAX_OID_LEN;
      asn_parse_objid (var_val, &len, &vp->type, objid, &vp->val_len);
      vp->val_len *= sizeof (oid);
      vp->val.objid = (oid *) calloc(1, (unsigned) vp->val_len);
      memmove (vp->val.objid, objid, vp->val_len);
      break;
    case SNMP_NOSUCHOBJECT:
    case SNMP_NOSUCHINSTANCE:
    case SNMP_ENDOFMIBVIEW:
    case ASN_NULL:
      break;
    case ASN_BIT_STR:
      vp->val.bitstring = (u_char *) calloc(1, vp->val_len);
      asn_parse_bitstring (var_val, &len, &vp->type,
			   vp->val.bitstring, &vp->val_len);
      break;
    default:
      sprintf(errmsg, "Error: bad type returned (%x)\n", vp->type);
	errmsg += strlen( errmsg );
      badtype = 1;
	return 0;
      break;
    }

  if (!netsnmp_ds_get_boolean
      (NETSNMP_DS_LIBRARY_ID, NETSNMP_DS_LIB_PRINT_FULL_OID))
    {
      netsnmp_ds_toggle_boolean (NETSNMP_DS_LIBRARY_ID,
				 NETSNMP_DS_LIB_PRINT_FULL_OID);
    }
  if (!netsnmp_ds_get_boolean
      (NETSNMP_DS_LIBRARY_ID, NETSNMP_DS_LIB_PRINT_NUMERIC_OIDS))
    {
      netsnmp_ds_toggle_boolean (NETSNMP_DS_LIBRARY_ID,
				 NETSNMP_DS_LIB_PRINT_NUMERIC_OIDS);
    }
  if (!netsnmp_ds_get_boolean
      (NETSNMP_DS_LIBRARY_ID, NETSNMP_DS_LIB_QUICK_PRINT))
    {
      netsnmp_ds_toggle_boolean (NETSNMP_DS_LIBRARY_ID,
				 NETSNMP_DS_LIB_QUICK_PRINT);
    }

  if (!strcmp (_docsis_snmp_label, "String"))	/* Strings need special treatment - see below  */
    {
      netsnmp_ds_toggle_boolean (NETSNMP_DS_LIBRARY_ID,
				 NETSNMP_DS_LIB_QUICK_PRINT);
      netsnmp_ds_toggle_boolean (NETSNMP_DS_LIBRARY_ID,
				 NETSNMP_DS_LIB_ESCAPE_QUOTES);
    }

  switch ((short) vp->type)
    {
    case ASN_OCTET_STR:
	if (str_isprint(vp->val.string, vp->val_len)) 
		{
		 	snprintf(out_buffer, omx, "\"%s\"", vp->val.string);
		} else { 
			snprint_hexadecimal (out_buffer, omx, vp->val.string, vp->val_len);
      			memset (_docsis_snmp_label, 0, 50);
      			sprintf (_docsis_snmp_label, "HexString");
		}
	break;

    case ASN_BIT_STR:
		snprint_hexadecimal (out_buffer, omx, vp->val.bitstring, vp->val_len);
		break;
 
    default: 
	snprint_value (out_buffer, omx, vp->name, vp->name_length, vp);

    }

    if ( subtree ) 
    {
   	enums = subtree->enums;	
	   	for (; enums; enums = enums->next) {
			if (enums->value == *vp->val.integer) {
			enum_string = enums->label;
       			break;
			}
		}
    }
	omx -= strlen( out_buffer );
	out_buffer += strlen( out_buffer );
	sprintf(out_buffer,";");
	omx--; out_buffer++;

  if (enum_string) 
	snprintf(out_buffer,omx, " /* %s */", enum_string);
	omx -= strlen( out_buffer );
	out_buffer += strlen( out_buffer );

  snmp_free_var (vp);

  return badtype;
}

unsigned int
encode_snmp_oid (char *oid_string, unsigned char *out_buffer,
	    unsigned int out_size)
{
  oid var_name[MAX_OID_LEN];
  size_t name_len = MAX_OID_LEN;
  unsigned char *data_ptr;

  if (!read_objid (oid_string, var_name, &name_len))
    {
      if (!get_node (oid_string, var_name, &name_len))
	{
	  sprintf(errmsg, "Can't find oid %s at line %d\n", oid_string, line);
		errmsg += strlen( errmsg );
	  return 0;
	}

    }

  data_ptr = asn_build_objid (out_buffer, &out_size,
			      (u_char) (ASN_UNIVERSAL | ASN_PRIMITIVE |
					ASN_OBJECT_ID), var_name, name_len);

  return data_ptr - out_buffer;
}


unsigned int
decode_snmp_oid (unsigned char *data, unsigned int data_len, char *out_buffer )
{
  oid this_oid[MAX_OID_LEN];
  size_t oid_len = MAX_OID_LEN;
  unsigned char type;
  void *retval;			/* generic pointer, will only use to check return value */
  int len;

  len = data_len;		/* The last char is the enable/disable switch */

  if ((retval =
       asn_parse_objid (data, &len, &type, this_oid, &oid_len)) == NULL)
    {
      sprintf(errmsg, "OID.parse.error");
	errmsg += strlen( errmsg );
      return 0;
    }
  else
    {
      netsnmp_ds_set_int (NETSNMP_DS_LIBRARY_ID,
			  NETSNMP_DS_LIB_OID_OUTPUT_FORMAT,
			  NETSNMP_OID_OUTPUT_NUMERIC);
      snprint_objid (out_buffer, omx, this_oid, oid_len);
    }
  return 1;
}



int hexadecimal_to_binary (const char *str, unsigned char * bufp) {
  int len, itmp;
  printf ("Hex string rx'd: %s\n", str);
  if (!bufp)
    return -1;
  if (*str && *str == '0' && (*(str + 1) == 'x' || *(str + 1) == 'X'))
    str += 2;
  for (len = 0; *str; str++)
    {
      if (isspace ((int) *str))
        continue;
      if (!isxdigit ((int) *str))
        return -1;
      len++;
      if (sscanf (str++, "%2x", &itmp) == 0)
        return -1;
      *bufp++ = itmp;
      if (!*str)
        return -1;              /* odd number of chars is an error */
    }
  return len;
}


/*
 * Print a string in hex format output ... for strings that are not alphanumeric, i.e.
 * HexString, BitString etc
 * str is the original string
 * str_len is the length of the original string
 * outbuf is a pointer to the output buffer
 */

void snprint_hexadecimal ( unsigned char *outbuf, size_t outsize, const char *str, size_t str_len )
{
  unsigned int i;
  unsigned char *cp;

  cp=outbuf;

  memset (outbuf, 0, outsize);
  sprintf(cp, "0x");
  cp = cp +2*sizeof(char);

  for (i=0; i<str_len && (unsigned int) (cp-outbuf) < outsize; i++) {
        sprintf(cp, "%02x", (unsigned char) str[i]);
        cp = cp +2*sizeof(char);
  }

}

