/* 
 *  DOCSIS configuration file encoder. 
 *  Copyright (c) 2001 Cornel Ciocirlan, ctrl@users.sourceforge.net.
 *  Copyright (c) 2002 Evvolve Media SRL,office@evvolve.com
 *  
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *  DOCSIS is a registered trademark of Cablelabs, http://www.cablelabs.com
 */

%{
#include <stdio.h>
#include <string.h>
#include <sys/types.h>

#include "docsis_symtable.h"
#include "docsis_encode.h"
#include "docsis_yy.tab.h"
#include "docsis_snmp.h"
#include "util_md5.h"
#include "util_hmac_md5.h"

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif
#if YYTEXT_POINTER
extern char *yytext;
#else
extern char yytext[];
#endif

extern unsigned int line; 	/* current line number, defined in a.l */
extern char *errmsg;

/* Global Variable */
struct tlv_list *global_tlvlist = NULL;
 
%}                  

%union { 	/* Token types */
	int intval;		/* For integers */
	unsigned int uintval;		/* For integers */
	symbol_type *symptr; 		/* For token identifiers */
	char *strval;		 	/* For strings */
	unsigned char *ustrval; 	/* For other any other types */
	unsigned int ip;		/* For IP Addresses */
	struct tlv_list *tlvlist; 	/* For for struct tlvlist pointers */
	struct tlv *tlvptr;		/* For struct tlv pointers; */
}

%token <symptr>  T_IDENTIFIER
%token <uintval> T_INTEGER
%token <symptr>  T_IDENT_COS
%token <symptr>  T_IDENT_BPI
%token <symptr>  T_IDENT_SNMPW
%token <symptr>  T_IDENT_SNMPSET
%token <symptr>  T_IDENT_GENERIC
%token <strval>  T_ETHERMASK
%token <strval>  T_LABEL_OID
%token <strval>  T_IP
%token <strval>  T_MAC
%token <strval>  T_MAIN
%token <strval>  T_STRING
%token <strval>  T_HEX_STRING

%token <uintval>  T_ASNTYPE_INT
%token <uintval>  T_ASNTYPE_UINT
%token <uintval>  T_ASNTYPE_GAUGE
%token <uintval>  T_ASNTYPE_TICKS 
%token <uintval>  T_ASNTYPE_IP 
%token <uintval>  T_ASNTYPE_OBJID 
%token <uintval>  T_ASNTYPE_STRING 
%token <uintval>  T_ASNTYPE_HEXSTR 
%token <uintval>  T_ASNTYPE_DECSTR 
%token <uintval>  T_ASNTYPE_BITSTR 
%token <uintval>  T_ASNTYPE_BIGINT 
%token <uintval>  T_ASNTYPE_UBIGINT 
%token <uintval>  T_ASNTYPE_FLOAT 
%token <uintval>  T_ASNTYPE_DOUBLE     
%token <uintval>  T_TLV_CODE
%token <uintval>  T_TLV_LENGTH
%token <uintval>  T_TLV_VALUE


%type <tlvptr>  assignment_stmt
%type <tlvlist> assignment_list
%type <tlvlist> config_stmt
/* %type <tlvlist> config_stmt_list */
/* %type <tlvlist> config_stmt */

%%
/* 

How this works ? 

When we recognize an assignment statement (for example "MaxCPE 13;") we create 
a "struct tlv" which contains the type associated with this configuration 
setting, the length of this configuration setting and the value. The type and
lenth are in docsis_symtable.h. Consecutive assignment_stmts are reduced  
to an assignment_list, creating a "tlvlist" which is essentially a list of 
pointers to stuct tlv's. 

"Special" cases like ClassOfService (configuration settings which have 
sub-types) are treated a bit different.

For example, within the ClassOfService statement we have a list of 
assignment_stmts which are reduced to an assignment_list, and then to a 
cos_stmt. When we flatten/reduce the cos_stmt to a config_stmt_list we merge the 
cos_stmt tlvlist with the tlvlist corresponding to config_stmt_list (may be
empty if cos_stmt is the first statement that appears in the config file) so we
have a single tlvlist. 

The idea is that, when we finish parsing, we end up with a "global_tlvlist" 
which contains pointers to all the struct tlv's that we assembled while parsing
the configuration file. At this point, we "flatten" this list into a tlvbuf, 
add CM MIC and CMTS MIC and pad and that's it. */


/* Definitions of bison/yacc grammar */

main_stmt: 	T_MAIN '{' assignment_list '}' {
			global_tlvlist = $3; }
		;

assignment_list: assignment_list assignment_stmt { $$ = add_tlv_to_list ($1,$2); } 
		| assignment_stmt {  $$=add_tlv_to_list(NULL,$1); } 
		| assignment_list config_stmt { 
			$$ = merge_tlvlist($1,$2);  }
		| config_stmt { 
			$$ = merge_tlvlist (NULL, $1); }
		; 

config_stmt:  	T_IDENTIFIER '{' assignment_list  '}'	{ 
			$$ = assemble_list_in_parent ( $1, $3 ); }
		;

assignment_stmt:  T_IDENTIFIER T_INTEGER ';' { 
			$$ = create_tlv ($1, (union t_val *)&$2);} 
		| T_IDENTIFIER T_STRING ';'  {
			$$ = create_tlv ($1, (union t_val *)&$2);}	
		| T_IDENTIFIER T_HEX_STRING ';'  {
			$$ = create_tlv ($1, (union t_val *)&$2);}	
		| T_IDENTIFIER T_IP ';' {
			$$ = create_tlv ($1, (union t_val *)&$2);}	
		| T_IDENTIFIER T_MAC ';' {
			$$ = create_tlv ($1, (union t_val *)&$2);}	
		| T_IDENTIFIER T_ETHERMASK ';' {
			$$ = create_tlv ($1, (union t_val *)&$2);}	
		| T_IDENTIFIER T_LABEL_OID ';' {
			$$ = create_tlv ($1, (union t_val *)&$2);}	
		| T_IDENT_SNMPW T_LABEL_OID T_INTEGER ';' {    
			$$ = create_snmpw_tlv ( $1, $2, (union t_val *) &$3 ); }
		| T_IDENT_SNMPSET T_LABEL_OID T_ASNTYPE_INT T_INTEGER ';' {
			$$ = create_snmpset_tlv($1,$2,'i',(union t_val *)&$4); }
		| T_IDENT_SNMPSET T_LABEL_OID T_ASNTYPE_GAUGE T_INTEGER ';' {
			$$ = create_snmpset_tlv($1,$2,'g',(union t_val *)&$4); }
	        | T_IDENT_SNMPSET T_LABEL_OID T_ASNTYPE_IP T_IP ';' {
			$$ = create_snmpset_tlv($1,$2,'a',(union t_val *)&$4); }
		| T_IDENT_SNMPSET T_LABEL_OID T_ASNTYPE_STRING T_STRING ';' {
			$$ = create_snmpset_tlv($1,$2,'s',(union t_val *)&$4); }
		| T_IDENT_SNMPSET T_LABEL_OID T_ASNTYPE_HEXSTR T_HEX_STRING ';' {
			$$ = create_snmpset_tlv($1,$2,'x',(union t_val *)&$4); }
		| T_IDENT_GENERIC T_TLV_CODE T_INTEGER T_TLV_LENGTH T_INTEGER T_TLV_VALUE T_HEX_STRING ';' {
			$$ = create_generic_tlv($1,$3,$5, (union t_val *)&$7); }
                ;                                        
%%

int yyerror(char *s) {
	sprintf(errmsg, "%d:%s token %s\n",line,s, yytext );
	errmsg += strlen(errmsg);
	sprintf(errmsg, "%d:%s token \n",line,s);
	errmsg += strlen(errmsg);
	return 0;
}


int  parse_input_buffer( char *inbuf, unsigned char *tlvbuf, int maxtlv ) {
	int	i = 0, cbuf = 0;
	int	bufsize = 0;
	char	*p;

	bufsize = strlen( inbuf );
	inbuf[bufsize++] = 0;
	inbuf[bufsize++] = 0;

#ifdef DEBUG
	yydebug = 1;
#endif
	yy_scan_buffer( inbuf, bufsize );
	yyparse();

	if (global_tlvlist == NULL) {
		sprintf( errmsg, "Global TLV List is NULL" );
		errmsg += strlen(errmsg);
		return 0;
	}

	for ( i=0; i < global_tlvlist->tlv_count; i++ ) {
		cbuf += 2; /* docs_code and len */
		cbuf += global_tlvlist->tlvlist[i]->tlv_len;
	}
	cbuf += 255;

	if (cbuf > maxtlv) {
		sprintf( errmsg, "TLV %d large than buffer %d", cbuf, maxtlv );
		errmsg += strlen(errmsg);
		return 0;
	}

	cbuf = flatten_tlvlist( tlvbuf, global_tlvlist );
	return( cbuf );
}
