/* 
 *  DOCSIS configuration file encoder. 
 *  Copyright (c) 2001 Cornel Ciocirlan, ctrl@users.sourceforge.net.
 *  Copyright (c) 2002,2003 Evvolve Media SRL,office@evvolve.com
 *  
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *  DOCSIS is a registered trademark of Cablelabs, http://www.cablelabs.com
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <stdio.h>
#include <errno.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <fcntl.h>

#include "libdocsis.h"
#include "docsis_symtable.h"
#include "docsis_decode.h"
#include "docsis_encode.h"
#include "docsis_yy.tab.h"
#include "docsis_snmp.h"
#include "util_md5.h"
#include "util_hmac_md5.h"

#include <net-snmp/net-snmp-config.h>
#include <net-snmp/utilities.h>
#include <net-snmp/config_api.h>
#include <net-snmp/output_api.h>
#include <net-snmp/mib_api.h>


char	*errmsg;	// If there is an error we write the error to the msg string


int  docsis_add_cm_mic( unsigned char *tlvbuf, int tlvlen, int maxtlv ) {
	MD5_CTX		mdContext;
	unsigned char	digest[16];

	if (tlvlen + 18 > maxtlv ) {
		sprintf( errmsg, "tlvlen %d exceded max %d", tlvlen, maxtlv );
		errmsg += strlen(errmsg);
		return tlvlen;
	}

	MD5Init( &mdContext );
	MD5Update( &mdContext, tlvbuf, tlvlen );
	MD5Final( digest, &mdContext );

	tlvbuf[tlvlen] = 6;
	tlvbuf[tlvlen + 1] = 16;
	memcpy ( tlvbuf + tlvlen + 2, digest, 16);

	return( tlvlen + 18 );
	/* we added the CM Message Integrity Check  */
}


int  docsis_add_cmts_mic( unsigned char *tlvbuf, int tlvlen, int maxtlv,
			  unsigned char *key,    int keylen ) {
	int 			i;
	register unsigned char	*cp, *dp;
	unsigned char		*cmts_tlvs;
	unsigned char		digest[17];

	if (tlvlen + 18 > maxtlv ) {
		sprintf( errmsg, "tlvlen %d exceded max %d", tlvlen, maxtlv );
		errmsg += strlen(errmsg);
		return tlvlen;
	}

	/* Only these configuration tlv's must be used to calculate the CMTS MIC */
#define NR_TLVS 17  
	unsigned char digest_order[NR_TLVS] =
		{ 1,2,3,4,17,43,6,18,19,20,22,23,24,25,28,29,26 }; 

	cmts_tlvs = (unsigned char *) calloc (1, tlvlen + 1 );
	dp = cmts_tlvs;

	for( i=0; i < NR_TLVS; i++ ) { 
		cp = tlvbuf;
		while ( (cp - tlvbuf) < tlvlen ) {
			if ( cp[0] == digest_order[i] ) {
				memcpy ( dp, cp, cp[1] + 2 );
				dp = dp + cp[1] + 2;
				cp = cp + cp[1] + 2;
			} else {
				cp = cp + cp[1] + 2;
			}
		}
  	}

	hmac_md5 ( cmts_tlvs, dp - cmts_tlvs, key, keylen, digest) ; 
	tlvbuf[tlvlen] = 7;		/* CMTS MIC */
	tlvbuf[tlvlen + 1] = 16;	/* length of MD5 digest */
	memcpy( &tlvbuf[tlvlen + 2], digest, 16);
	free( cmts_tlvs );

	return( tlvlen + 18 );
}



int  docsis_add_eod_and_pad( unsigned char *tlvbuf, int tlvlen, int maxtlv ) {
	int	nr_pads;

	if (tlvlen + 18 > maxtlv ) {
		sprintf( errmsg, "tlvlen %d exceded max %d", tlvlen, maxtlv );
		errmsg += strlen(errmsg);
		return tlvlen;
	}

	tlvbuf[tlvlen] = 255;
	tlvlen++;
	nr_pads = ((tlvlen >> 2) + 1 << 2) - tlvlen;

	memset( &tlvbuf[tlvlen], 0, nr_pads );

	return( tlvlen + nr_pads);
}



// encodes part of a config file. No Main needed.
int docsis_encode_part( char *inbuf, unsigned char *tlvbuf, int maxtlv ) {
	int	inlen, retval;
	char	*tmpstr;

	inlen = strlen( inbuf );
	tmpstr = (char *) calloc(1, inlen + 30 );

	strcpy( tmpstr, "Main {\n" );
	strcat( tmpstr, inbuf );
	strcat( tmpstr, "\n}\n" );

	retval = parse_input_buffer( tmpstr, tlvbuf, maxtlv );

	free( tmpstr );
	return( retval );
}

// encodes a complete config file. Full syntax needed.
int docsis_encode_whole( char *inbuf, unsigned char *tlvbuf, int maxtlv ) {
	return( parse_input_buffer( inbuf, tlvbuf, maxtlv ) );
}


int docsis_decode( unsigned char *in_buf, int in_len,
		   char *outbuf, int maxbuf ) {
	return( decode_buffer( in_buf, in_len, outbuf, maxbuf ) );
}



// init structures
void docsis_initialize( char *error_msgs ) {

	errmsg = error_msgs;

	setenv("MIBS", "ALL", 1);
	init_snmp("snmpapp");
	snmp_set_mib_warnings(2);

	if (! netsnmp_ds_get_boolean (NETSNMP_DS_LIBRARY_ID,  NETSNMP_DS_LIB_PRINT_NUMERIC_OIDS)) {
		netsnmp_ds_toggle_boolean(NETSNMP_DS_LIBRARY_ID, NETSNMP_DS_LIB_PRINT_NUMERIC_OIDS);
	} /* we want OIDs to appear in numeric form */

	if (! netsnmp_ds_get_boolean (NETSNMP_DS_LIBRARY_ID,  NETSNMP_DS_LIB_PRINT_NUMERIC_ENUM)) {
		netsnmp_ds_toggle_boolean(NETSNMP_DS_LIBRARY_ID, NETSNMP_DS_LIB_PRINT_NUMERIC_ENUM);
	} /* we want enums to appear in numeric form as integers */

	if (! netsnmp_ds_get_boolean (NETSNMP_DS_LIBRARY_ID,  NETSNMP_DS_LIB_PRINT_FULL_OID)) {
		netsnmp_ds_toggle_boolean(NETSNMP_DS_LIBRARY_ID, NETSNMP_DS_LIB_PRINT_FULL_OID);
	} /* we want to full numeric OID to be printed, including prefix */

//	if (! netsnmp_ds_get_boolean (NETSNMP_DS_LIBRARY_ID,  NETSNMP_DS_LIB_QUICK_PRINT)) {
//		netsnmp_ds_toggle_boolean(NETSNMP_DS_LIBRARY_ID, NETSNMP_DS_LIB_QUICK_PRINT);
//	} /* quick print - doesnt give you the type however */

//	if (! netsnmp_ds_get_boolean (NETSNMP_DS_LIBRARY_ID,  NETSNMP_DS_LIB_DONT_BREAKDOWN_OIDS)) {
//		netsnmp_ds_toggle_boolean(NETSNMP_DS_LIBRARY_ID, NETSNMP_DS_LIB_DONT_BREAKDOWN_OIDS);
//	}

//	if (! netsnmp_ds_get_boolean (NETSNMP_DS_LIBRARY_ID,  NETSNMP_DS_LIB_PRINT_UCD_STYLE_OID)) {
//		netsnmp_ds_toggle_boolean(NETSNMP_DS_LIBRARY_ID, NETSNMP_DS_LIB_PRINT_UCD_STYLE_OID);
//	}

//	if (! netsnmp_ds_get_boolean (NETSNMP_DS_LIBRARY_ID,  NETSNMP_DS_LIB_DONT_PRINT_UNITS)) {
//		netsnmp_ds_toggle_boolean(NETSNMP_DS_LIBRARY_ID, NETSNMP_DS_LIB_DONT_PRINT_UNITS);
//	}

	netsnmp_ds_set_int(NETSNMP_DS_LIBRARY_ID, NETSNMP_DS_LIB_OID_OUTPUT_FORMAT,
			NETSNMP_OID_OUTPUT_SUFFIX); 

	if (! netsnmp_ds_get_boolean (NETSNMP_DS_LIBRARY_ID, NETSNMP_DS_LIB_RANDOM_ACCESS)) {
		netsnmp_ds_toggle_boolean(NETSNMP_DS_LIBRARY_ID, NETSNMP_DS_LIB_RANDOM_ACCESS);
	}  /* so we can use sysContact.0 instead of system.sysContact.0  */
}


