
/* 
 *  DOCSIS configuration file encoder. 
 *  Copyright (c) 2001 Cornel Ciocirlan, ctrl@users.sourceforge.net.
 *  Copyright (c) 2002 Evvolve Media SRL,office@evvolve.com
 *  
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *  DOCSIS is a registered trademark of Cablelabs, http://www.cablelabs.com
 */




#ifndef _LIBDOCSIS_H
#define _LIBDOCSIS_H

/* Functions exported by libdocsis */

// docsis_initialize()
// init structures and setup return messages
// *****************************************
// error_msgs -- should be a 1K string to hold
//               any errors.
void docsis_initialize( char *error_msgs );


//***************************************
//***************************************
// Encoding Functions


// int = docsis_encode_part()
// encodes part of a config file. No Main needed.
// **********************************************
// inbuf  -- pointer to config file string
// tlvbuf -- pointer to buffer binary tlv data
// maxtlv -- maximum size of tlv buffer
// returns the size of the data in the tlv buffer.
int  docsis_encode_part( char *inbuf, unsigned char *tlvbuf, int maxtlv );


// int = docsis_encode_whole()
// encodes a complete config file. Full syntax needed.
// *****************************************************
// inbuf  -- pointer to config file string
// tlvbuf -- pointer to buffer binary tlv data
// maxtlv -- maximum size of tlv buffer
// returns the size of the data in the tlv buffer.
int  docsis_encode_whole( char *inbuf, unsigned char *tlvbuf, int maxtlv );

// int = docsis_add_cm_mic()
// Adds the CM Message Integrity Check
// ****************************************************
// tlvbuf -- pointer to buffer binary tlv data
// tlvlen -- current length of tlv data
// maxtlv -- maximum size of tlv buffer
// returns the size of the data in the tlv buffer.
int  docsis_add_cm_mic( unsigned char *tlvbuf, int tlvlen, int maxtlv );

// int = docsis_add_cmts_mic()
// Adds the CMTS Message Integrity Check
// tlvbuf -- pointer to buffer binary tlv data
// tlvlen -- current length of tlv data
// maxtlv -- maximum size of tlv buffer
// key    -- pointer to key string
// keylen -- length of key
// returns the size of the data in the tlv buffer.
int  docsis_add_cmts_mic( unsigned char *tlvbuf, int tlvlen, int maxtlv,
			  unsigned char *key,    int keylen );

// int = docsis_add_eod_and_pad()
// Finish up the file.
// tlvbuf -- pointer to buffer binary tlv data
// tlvlen -- current length of tlv data
// maxtlv -- maximum size of tlv buffer
// returns the size of the data in the tlv buffer.
int  docsis_add_eod_and_pad( unsigned char *tlvbuf, int tlvlen, int maxtlv );


//***************************************
//***************************************
// Decoding Functions

// int = docsis_decode()
// Decode a binary tlv config file into human readable config file
// in_tlv  -- pointer to binary tlv data buffer
// in_len  -- length of tlv data
// outbuf  -- pointer to string for output
// maxbuf  -- maximum size of the output buffer
// returns the size of the data in the output buffer
int docsis_decode( unsigned char *in_tlv, int in_len,
		    char *outbuf, int maxbuf );

#endif /* _LIBDOCSIS_H */

