/* cfgen.c
 *
 * docsis_server
 *
 * Copyright (C) 2002 Access Communications
 *                    email docsis_guy@accesscomm.ca
 *
 * Designed for use with DOCSIS cable modems and hosts.
 * And for use with MySQL Database Server.
 *
 ********************************************************************
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <pwd.h>
#include <fcntl.h>
#include <string.h>
#include <stdlib.h>
#include <sys/wait.h>
#include <sys/stat.h>
#include <sys/ioctl.h>
#include <arpa/inet.h>
#include <netdb.h>
#include <netinet/in.h>
#include <net/if.h>
#include <linux/sockios.h>
#include <stdio.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <unistd.h>
#include <syslog.h>
#include <signal.h>
#include <errno.h>
#include <time.h>

#define CF_GEN	1

#include "../src-mysql/my_sql.h"
#include "../src-config/config_file.h"
#include "../Version.h"

#define MAX_CONFIGS_IN_MEM	5000

typedef struct cfg_cache {
	u_int32_t	ipaddr;
	time_t		stale;
	u_int32_t	size;
	u_int8_t	*buf;
} CfgCache;



// Global Variables
extern int	dhcp2cfgen[2];
extern int	tftp2cfgen[2];
extern int	cfgen2tftp[2];
u_int32_t	ipaddr[2];
CfgCache	cfga[MAX_CONFIGS_IN_MEM];
int		cfg_cnt;


// Prototypes
int	my_generate_config_file( void );
int	find_config( void );
void	drop_config( int id );


void cf_generator( void ) {
	char			*euid_pref, *tftp_dir, *dbhost_pref;
	struct passwd		*pass_struct;
	struct stat		tftp_stat;
	long			file_flags;
	int			id, bytes1, bytes2, maxfd;
	fd_set			rfds;
	struct timeval		tv;

	InitSQLconnection();
	my_LoadChecker();

	tftp_dir = my_GetTFTPdir();
	if ( stat( tftp_dir, &tftp_stat ) != 0 ) {
		fprintf(stderr,"Could not read tftp dir: %s", tftp_dir );
		exit(1);
	}
	chdir( tftp_dir );

	dbhost_pref = GetConfigVar( "mysql-host" );
	euid_pref = GetConfigVar( "effective-userid" );
	pass_struct = getpwnam( euid_pref );

	if (strcmp(dbhost_pref,"localhost")) {
		// We cannot chroot if connect to the DB via
		// a unix socket vs tcp
		chroot( tftp_dir );
	}
	if (pass_struct) {
		seteuid( pass_struct->pw_uid );
		setegid( pass_struct->pw_gid );
	}

	my_syslog( MLOG_DHCP, "docsis_server CF-GEN version %s activated", VERSION);

	file_flags = fcntl( dhcp2cfgen[0], F_GETFL );
	file_flags |= O_NONBLOCK;
	fcntl( dhcp2cfgen[0], F_SETFL, file_flags );

	file_flags = fcntl( tftp2cfgen[0], F_GETFL );
	file_flags |= O_NONBLOCK;
	fcntl( tftp2cfgen[0], F_SETFL, file_flags );

	maxfd = dhcp2cfgen[0];
	if (tftp2cfgen[0] > maxfd) maxfd = tftp2cfgen[0];
	maxfd++;

	cfg_cnt = 0;
	for (id=0; id < MAX_CONFIGS_IN_MEM; id++ ) {
		cfga[id].ipaddr = 0;
		cfga[id].stale = 0;
		cfga[id].size = 0;
		cfga[id].buf = NULL;
	}

	my_CONFIG_Load();
	while (1) {
		FD_ZERO( &rfds );
		FD_SET( dhcp2cfgen[0], &rfds );
		FD_SET( tftp2cfgen[0], &rfds );
		tv.tv_sec = 15;
		tv.tv_usec = 0;
		bytes1 = 0;
		bytes2 = 0;

		select( maxfd, &rfds, NULL, NULL, &tv );

		if ( FD_ISSET( dhcp2cfgen[0], &rfds ) ) {
			bytes1 = read( dhcp2cfgen[0], &ipaddr, 8 );

			if (bytes1 > 0) {
				if (bytes1 == 8) {
					my_generate_config_file();
				}
			}
		}

		if ( FD_ISSET( tftp2cfgen[0], &rfds ) ) {
			bytes2 = read( tftp2cfgen[0], &ipaddr, 4 );

			if (bytes2 > 0) {
				if (bytes2 == 4) {
					id = find_config();
					if (id >= 0) {
						write( cfgen2tftp[1], &(cfga[id].size), 4 );
						write( cfgen2tftp[1], cfga[id].buf, cfga[id].size );
						drop_config( id );
					} else {
						u_int8_t tmp_buf[20000];
						u_int32_t bytes;
						bytes = my_ReverseNet_CM( ntohl( ipaddr[0] ) );
						ipaddr[1] = htonl( bytes );
						bytes = gen_Config_File( ipaddr[1], tmp_buf, 20000 );
						if (bytes == 0) {
							ipaddr[0] = 0;
							write( cfgen2tftp[1], ipaddr, 4 );
						} else {
							write( cfgen2tftp[1], &bytes, 4 );
							write( cfgen2tftp[1], tmp_buf, bytes );
						}
					}
				}
			}
		}

		if (bytes1 == 0 && bytes2 == 0) {
			my_CONFIG_Load();
		}
	}
}


int my_generate_config_file( void ) {
	static u_int8_t	tmp_buf[20000];
	int		id, bytes;
	time_t		cur_t;
	int		ofd;
	static char	*dynamic_debug_pref;
	char		outfile[100];
	struct in_addr          s_ip;

	cur_t = time(NULL);

	bytes = gen_Config_File( ipaddr[0], tmp_buf, 20000 );

	if (bytes == 0) return 0;

	if (dynamic_debug_pref == NULL) {
		dynamic_debug_pref = GetConfigVar( "dynamic-config-debug" );
	}

	if (*dynamic_debug_pref == 'y') {
		char    x1[20], x2[20];
		s_ip.s_addr = ipaddr[0]; strcpy( x1, inet_ntoa( s_ip ) );
		s_ip.s_addr = ipaddr[1]; strcpy( x2, inet_ntoa( s_ip ) );
		sprintf( outfile, "dynamic/dynamic--%s--%s.cfg", x1, x2 );
		if ( (ofd = open( outfile, O_CREAT | O_WRONLY, 0644 ))==-1 )  {
			fprintf(stderr,"Error opening file %s: %s",outfile,strerror(errno));
		}
		write( ofd, tmp_buf, bytes );
		close( ofd );
	}

	for (id=0; id < MAX_CONFIGS_IN_MEM; id++ ) {
		if (cfga[id].ipaddr != 0 && cur_t > cfga[id].stale) drop_config( id );
		if (cfga[id].ipaddr == 0) {
			cfg_cnt++;
			cfga[id].stale = cur_t + 180;
			cfga[id].ipaddr = ipaddr[1];
			cfga[id].size = bytes;
			cfga[id].buf = calloc(1, bytes + (bytes % 4) );
			memcpy( cfga[id].buf, tmp_buf, bytes );
			return 1;
		}
	}

	drop_config( 0 );

	cfg_cnt++;
	cfga[0].stale = cur_t + 180;
	cfga[0].ipaddr = ipaddr[1];
	cfga[0].size = bytes;
	cfga[0].buf = calloc(1, bytes + (bytes % 4) );
	memcpy( cfga[0].buf, tmp_buf, bytes );
	return 1;
}


int find_config( void ) {
	int	id;

	for (id=0; id < MAX_CONFIGS_IN_MEM; id++ ) {
		if (cfga[id].ipaddr == ipaddr[0] ) return id;
	}
	return -1;
}


void	drop_config( int id ) {
	if (cfga[id].ipaddr == 0) return;

	cfg_cnt--;

	cfga[id].ipaddr = 0;
	cfga[id].stale = 0;
	cfga[id].size = 0;
	free( cfga[id].buf );
	cfga[id].buf = NULL;
}
