/* my_sql.c
 *
 * docsis_server
 *
 * Copyright (C) 2002 Access Communications
 *                    email docsis_guy@accesscomm.ca
 *
 * Designed for use with DOCSIS cable modems and hosts.
 * And for use with MySQL Database Server.
 *
 ********************************************************************
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <string.h>
#include <stdlib.h>
#include <sys/wait.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <arpa/inet.h>
#include <netdb.h>
#include <netinet/in.h>
#include <net/if.h>
#include <linux/sockios.h>
#include <stdio.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <unistd.h>
#include <fcntl.h>
#include <time.h>
#include <syslog.h>
#include <signal.h>
#include <errno.h>

#include <stdarg.h>
#include <mysql/mysql.h>
#include <mysql/mysql_com.h>
#include <mysql/mysql_version.h>

#define DHCP_SERVER
#define TFTP_SERVER
#define SYSLOG_SERVER
#define CF_GEN
#define BACKG

#include "../src-dhcpd/dhcpd.h"
#include "../src-dhcpd/options.h"
#include "my_sql.h"
#include "../src-config/config_file.h"
#include "../docsis_lib/libdocsis.h"

/* Global Variables */
MYSQL		my_sql, *my_sock;
static float	my_LoadAvg = 0.0;
static float	my_MaxLoad = 0.0;


void  InitSQLconnection() {
	int	count;
	char	*dt, *tim;

	mysql_init( &my_sql );

	get_date_time(&dt,&tim);

	for( count=0; count<10; count++ ) {
		my_sock = mysql_real_connect( &my_sql,
			GetConfigVar( "mysql-host" ),
			GetConfigVar( "mysql-user" ),
			GetConfigVar( "mysql-password" ),
			GetConfigVar( "mysql-dhcp-database" ), 0, NULL, 0);
		if (my_sock) { return; }
		fprintf( stderr, "Can not connect to database server (attempt = %d)\n\n", (count + 1) );
		if (count < 9) sleep( 30 );
	}

	fprintf( stderr, "Can not connect to database server\n\n" );

	fprintf( stderr, "Config File: Host = '%s' \n", GetConfigVar( "mysql-host" ) );
	fprintf( stderr, "Config File: User = '%s' \n", GetConfigVar( "mysql-user" ) );
	fprintf( stderr, "Config File: Password = '%s' \n", GetConfigVar( "mysql-password" ) );
	fprintf( stderr, "Config File: Dhcp-Database = '%s' \n", GetConfigVar( "mysql-dhcp-database" ) );
	fprintf( stderr, "\n\n" );
}

void	CloseSQLconnection() {
	mysql_close( my_sock );
}

/* ****************************************************************** */
/* ****************************************************************** */
/* Various SQL routines */

void	my_syslog( const int priority, const char * fmt, ... ) {
	va_list		list;
	static char	mbuf[1024 + 1];
	char		*p = NULL, *dt = NULL, *tim = NULL;
	char		qbuf[500], qval[500];
	static char	*buf_pref = NULL;

	bzero( mbuf, 1024 );
	bzero( qbuf, 499 );
	bzero( qval, 499 );

	va_start (list, fmt);
	vsnprintf (mbuf, sizeof mbuf, fmt, list);
	va_end (list);

	for( p = mbuf; *p != '\0'; p++) {
		if (*p == 34) { *p = '`'; }
		if (*p == 39) { *p = '`'; }
		if (*p == 92) { *p = '/'; }
	}
	*(mbuf + 200) = 0;

	get_date_time( &dt, &tim );

	strcpy( qbuf, "insert delayed into sys_log " );
	strcat( qbuf, "( thedate, thetime, priority, message ) values " );
	sprintf( qval, "( '%s', '%s', %d, '%s' ) ", dt, tim, priority, mbuf );
	strcat( qbuf, qval );

	if (buf_pref == NULL) {
		buf_pref = GetConfigVar( "suspend-system-logging" );
		if (*buf_pref == ' ') { buf_pref = "yes"; }
	}

	Save_SQL_Update( buf_pref, qbuf, CHECK_LOAD_NO );
}

/* ****************************************************************** */

void  my_dhcplog( int sending, dhcp_message *message ) {
	char	qbuf[900], qbuf2[900];
	char	*s_send, *s_type, *s_dynam, *s_mcache, *dt, *tim;
	static char	*buf_pref = NULL;
	u_int32_t	ipaddr;

	ipaddr = message->b_ipaddr;

	s_send = "ACK";  /* sending = DHCP_ACK */
	if (sending == DHCP_OFFER) { s_send = "OFFER"; } else
	if (sending == DHCP_NAK)   { s_send = "NAK";   } else
	if (sending == DHCP_Q_ACK) { s_send = "Q_ACK"; } else
	if (sending == DHCP_Q_NAK) { s_send = "Q_NAK"; }

	if (*s_send == 'Q' && ipaddr == 0) { ipaddr = ntohl( message->in_pack.ciaddr ); }

	s_type = "CM";   /* lease_type = LEASE_CM */
	if (message->lease_type == LEASE_CPE ) { s_type = "CPE"; }
	if (message->lease_type == LEASE_MTA ) { s_type = "MTA"; }
	if (message->lease_type == LEASE_NOAUTH ) { s_type = "NO-AUTH"; }

	s_dynam = "NO";
	if (message->cpe_type == CPE_DYNAMIC ) { s_dynam = "YES"; }

	s_mcache = "NO";   /* mcache = 0 */
	if (message->mcache ) { s_mcache = "YES"; }

	get_date_time( &dt, &tim );

	strcpy( qbuf, "insert delayed into dhcp_log " );
	strcat( qbuf, "( thedate, thetime, sending, type, dynamic_flag, mcache_flag, ");
	strcat( qbuf, "b_macaddr, b_ipaddr, b_modem_macaddr, subnum ) " );
	strcat( qbuf, " values ( '%s', '%s', '%s', '%s', '%s', '%s', ");
	strcat( qbuf, "'%llu', '%lu', '%llu', '%llu' )" );

	sprintf( qbuf2, qbuf, dt, tim, s_send, s_type, s_dynam, s_mcache,
		message->b_macaddr, ipaddr,
		message->b_modem_macaddr, message->subnum );

	if (buf_pref == NULL) {
		buf_pref = GetConfigVar( "suspend-dhcp-logging" );
		if (*buf_pref == ' ') { buf_pref = "yes"; }
	}

	Save_SQL_Update( buf_pref, qbuf2, CHECK_LOAD_YES );
}

/* ****************************************************************** */

void  my_LoadChecker( void ) {
	static int	fd = 0;
	char		buffer[1024], *loadchr, *p;
	float		avg;
	int		count;
	time_t		new_time;
	static time_t	old_time = 0;

	if (fd == 0) {
		fd = open ("/proc/loadavg", O_RDONLY);
	}
	if (fd == -1) { my_LoadAvg = 0; return; }

	new_time = time( NULL );
	if ( (new_time - old_time) > 15 ) {	// Check every 15 seconds

		loadchr = GetConfigVar( "suspend-load-avg" );
		if (*loadchr == ' ') { loadchr = "1.0"; }
		my_MaxLoad = atof( loadchr );

		bzero( buffer, 1024 );
		lseek( fd, 0, SEEK_SET );
		count = read (fd, buffer, sizeof(buffer));
		if (count <= 0) { my_LoadAvg = 0; return; }
		p = buffer;  while( *p != 0 && *p != ' ') { p++; }
		*p = 0;
		avg = atof( buffer );
		if (avg > my_MaxLoad && my_LoadAvg < my_MaxLoad) {
			my_syslog( MLOG_ERROR, "Load Average is too high. Suspending Logging. %3.3f", avg );
		}
		if (avg < my_MaxLoad && my_LoadAvg > my_MaxLoad) {
			my_syslog( MLOG_ERROR, "Load Average is OK. %3.3f", avg );
		}
		my_LoadAvg = avg;
		old_time = new_time;
	}
}


/* ****************************************************************** */


/* include SQL functions for the DHCP Memory Cache functions */
/*	void  my_MCACHE_Load( void );
	int   my_findMAC_CACHE( dhcp_message *message );
*/
#include "my_sql_mcache.1"


/* ****************************************************************** */

int   my_findMAC_CM( dhcp_message *message );
int   my_findMAC_CPE( dhcp_message *message );
void  my_findCM_MAC_CACHE( dhcp_message *message );

/* return 1 if found. return 0 if not found */
int   my_findMAC( dhcp_message *message ) {
	int	result;

	/* If we become busy use a special cached version of the cablemodems */
	result = my_findMAC_CACHE( message );
	if (result) {
		return result;
	}

	result = my_findMAC_CM( message );
	if (result) {
		return LEASE_CM;
	}

	result = my_findMAC_CPE( message );
	if (result) {
		return LEASE_CPE;
	}
	return LEASE_NOT_FOUND;
}


int   my_findMAC_CM( dhcp_message *message ) {
	char		qbuf[1500];
	MYSQL_RES	*res;
	MYSQL_ROW	myrow;
	struct in_addr	inp;
	int		numr;

	strcpy( qbuf, "select ipaddr, config_file, cmts_vlan, subnum ");
	strcat( qbuf, "cfg_static_ip, cfg_dynamic_ip ");
	strcat( qbuf, "from docsis_modem where modem_macaddr = '" );
	strcat( qbuf, message->s_macaddr );
	strcat( qbuf, "' " );

	if ( mysql_ping( my_sock ) ) {
		fprintf( stderr, "no sql server -- %s\n", mysql_error( my_sock ) );
		return 0;
	}
	if ( mysql_query( my_sock, qbuf ) )  {
		fprintf( stderr, "sql error\nq=%s \ne=%s\n", qbuf,
			mysql_error( my_sock ) );
		return 0;
	}
	res = mysql_store_result( my_sock );
	if (res == NULL) { return 0; }

	numr = mysql_num_rows( res );
	if (numr == 0) { mysql_free_result(res); return 0; }

	myrow = mysql_fetch_row( res );
	if (myrow == NULL) {
		fprintf( stderr, "sql no_fetch_row\nq=%s \ne=%s\n", qbuf,
			mysql_error( my_sock ) );
		mysql_free_result(res);
		return 0;
	}

	message->lease_type = LEASE_CM;

	strncpy( message->s_ipaddr, (myrow)[0], 20 );
	inet_aton( (myrow)[0], &inp );
	message->ipaddr = inp.s_addr;
	message->b_ipaddr = ntohl( message->ipaddr );

	strncpy( message->cfname, (myrow)[1], MAX_CONFIG_FILE_NAME );

	message->vlan        = atoi(  (myrow)[2] );
	message->subnum      = atoll( (myrow)[3] );
	message->static_ips  = atoi(  (myrow)[4] );
	message->dynamic_ips = atoi(  (myrow)[5] );

	mysql_free_result(res);
	return 1;
}


int   my_findMAC_CPE( dhcp_message *message ) {
	char		qbuf[1500];
	MYSQL_RES	*res;
	MYSQL_ROW	myrow;
	struct in_addr	inp;
	int		numr;

	strcpy( qbuf, "select ipaddr, dynamic_flag, lockip_flag, ");
	strcat( qbuf, "modem_macaddr, cmts_vlan, ");
	strcat( qbuf, "unix_timestamp(end_time) - unix_timestamp() ");
	strcat( qbuf, "from dhcp_leases where macaddr = '" );
	strcat( qbuf, message->s_macaddr );
	strcat( qbuf, "' order by ipaddr limit 1" );

	if ( mysql_ping( my_sock ) ) {
		fprintf( stderr, "no sql server -- %s\n", mysql_error( my_sock ) );
		return 0;
	}
	if ( mysql_query( my_sock, qbuf ) )  {
		fprintf( stderr, "sql error\nq=%s \ne=%s\n", qbuf,
			mysql_error( my_sock ) );
		return 0;
	}
	res = mysql_store_result( my_sock );
	if (res == NULL) { return 0; }

	numr = mysql_num_rows( res );
	if (numr == 0) { mysql_free_result(res); return 0; }

	myrow = mysql_fetch_row( res );
	if (myrow == NULL) {
		fprintf( stderr, "sql no_fetch_row\nq=%s \ne=%s\n", qbuf,
			mysql_error( my_sock ) );
		mysql_free_result(res);
		return 0;
	}

	message->lease_type = LEASE_CPE;

	strncpy( message->s_ipaddr, (myrow)[0], 20 );
	inet_aton( (myrow)[0], &inp );
	message->ipaddr = inp.s_addr;
	message->b_ipaddr = ntohl( message->ipaddr );

	if ( strcmp( (myrow)[1],"YES" ) == 0 ) {
		message->cpe_type = CPE_DYNAMIC;
	}

	if ( strcmp( (myrow)[2],"YES" ) == 0 ) {
		message->lockip = 1;
	}

	if (message->b_modem_macaddr == 0) {
		strcpy( message->s_modem_macaddr, (myrow)[3] );
		cv_addrmac( message->modem_macaddr, message->s_modem_macaddr, 6 );
		cv_mac2num( &(message->b_modem_macaddr), message->modem_macaddr );
	}

	message->vlan = atoi( (myrow)[4] );

	message->lease_time = atol( (myrow)[5] );

	mysql_free_result(res);

	my_findCM_MAC_CACHE( message );
	return 1;
}


/* ****************************************************************** */


int   my_findIP_QUERY( char *qbuf, dhcp_message *message );

int   my_findIP( dhcp_message *message, int find_type ) {
	char		qbuf[500], *ip;
	int		res;
	struct in_addr	inp;

	ip = message->s_ipaddr; /* find_type = IPADDR */
	if (find_type == FIND_CIADDR) { ip = message->in_ciaddr; } else
	if (find_type == FIND_YIADDR) { ip = message->in_yiaddr; }

	inet_aton( ip, &inp );
	message->b_ipaddr = ntohl( inp.s_addr );
	strcpy( message->s_ipaddr, ip );
	ignore_repeated_customers( message );
	if (message->lease_type == LEASE_REJECT) return 0;

	strcpy( qbuf, "select macaddr, subnum, dynamic_flag, modem_macaddr, " );
	strcat( qbuf, "unix_timestamp(end_time) - unix_timestamp() ");
	strcat( qbuf, "from dhcp_leases where ipaddr = '" );
	strcat( qbuf, ip );
	strcat( qbuf, "' " );

	res = my_findIP_QUERY( qbuf, message );
	if ( res ) {
		message->lease_type = LEASE_CPE;
		return 1;
	}

	strcpy( qbuf, "select modem_macaddr, subnum, 'NO', modem_macaddr, 1 as lease_time from docsis_modem where ipaddr = '" );
	strcat( qbuf, ip );
	strcat( qbuf, "' " );

	res = my_findIP_QUERY( qbuf, message );
	if ( res ) {
		message->lease_type = LEASE_CM;
		return 1;
	}

	return 0;
}

int   my_findIP_QUERY( char *qbuf, dhcp_message *message ) {
	MYSQL_RES	*res;
	MYSQL_ROW	myrow;

	if ( mysql_ping( my_sock ) ) {
		fprintf( stderr, "no sql server -- %s\n", mysql_error( my_sock ) );
		return 0;
	}
	if ( mysql_query( my_sock, qbuf ) ) {
		fprintf( stderr, "sql error\nq=%s \ne=%s\n", qbuf,
			mysql_error( my_sock ) );
		return 0;
	}
	res = mysql_store_result( my_sock );
	if (res == NULL) return 0;

	myrow = mysql_fetch_row( res );
	if (myrow == NULL) { mysql_free_result(res); return 0; }

	strncpy( message->s_macaddr, (myrow)[0], 20 );
	cv_addrmac( message->macaddr, message->s_macaddr, 6 );
	cv_mac2num( &(message->b_macaddr), message->macaddr );
	if (message->b_macaddr == 0) { mysql_free_result(res); return 0; }

	message->subnum = atoll( (myrow)[1] );

	if ( strcmp( (myrow)[2],"YES" ) == 0 ) {
		message->cpe_type = CPE_DYNAMIC;
	}

	if (message->b_modem_macaddr == 0) {
		strcpy( message->s_modem_macaddr, (myrow)[3] );
		cv_addrmac( message->modem_macaddr, message->s_modem_macaddr, 6 );
		cv_mac2num( &(message->b_modem_macaddr), message->modem_macaddr );
		if (message->b_modem_macaddr == 0)	// 123456789012
			strcpy( message->s_modem_macaddr, "000000000000" );
	}

	message->lease_time = atol( (myrow)[4] );

	mysql_free_result(res);

	my_findCM_MAC_CACHE( message );
	return 1;
}


/* ****************************************************************** */


/* include SQL functions for the Network and Options functions */
/*	void  my_Clear_Opts( int opt_id );
	void  my_Clear_Nets( void );
	int   my_Load_Opts( int opt_id );
	void  my_Load_Nets( void );
	ConfigOptsPtr my_Get_Opt( int opt_id );
	int   my_Get_Net( dhcp_message *message );
*/
#include "my_sql_subnets.1"


/* ****************************************************************** */

int my_CheckIP( u_int32_t ipaddr );
void my_Count_Allowed_IPs( dhcp_message *message );
void my_Count_Used_IPs( dhcp_message *message );

void my_getNewIP( dhcp_message *message ) {
	int		sn;
	int		ip_stat = 0, ip_dyn = 0, lease_type = 0, cpe_type = 0;
	u_int32_t	n_ord, h_min, h_max, ss;
	char		*alert_pgm;
	struct in_addr	inp;
	static time_t	old_time = 0;
	static time_t	old_time2 = 0;
	static char	*ip_pref = NULL;
	time_t		new_time = 0;

	if (num_nets == 0) {
		my_Load_Nets();
		old_time = old_time2 = time( NULL );
	}

	if (ip_pref == NULL) {
		ip_pref = GetConfigVar( "ip-http-mgmt" );
	}

	my_Count_Allowed_IPs( message );
	my_Count_Used_IPs( message );

	lease_type = LEASE_NOAUTH;
	ip_stat = message->static_ips - message->static_cur;
	if (ip_stat > 0) {
		lease_type = LEASE_CPE;
		cpe_type = CPE_STATIC;
	} else {
		ip_dyn  = message->dynamic_ips - message->dynamic_cur;
		if (ip_dyn > 0) {
			lease_type = LEASE_CPE;
			cpe_type = CPE_DYNAMIC;
		}
	}
	if (*ip_pref != 'y') {
		if (lease_type == LEASE_NOAUTH) {
			lease_type = LEASE_CPE;
			if (message->static_ips > 0) {
				cpe_type = CPE_STATIC;
			} else {
				cpe_type = CPE_DYNAMIC;
			}
		}
	}

	for( sn=0; sn < num_nets; sn++ ) {
		if (my_nets[sn]->cmts_ip != message->in_pack.giaddr) { continue; }
		if (my_nets[sn]->grant_flag == 0) { continue; }
		if (my_nets[sn]->full_flag == 1) { continue; }
		if (my_nets[sn]->nettype != lease_type) { continue; }
		if (my_nets[sn]->dynamic_flag != cpe_type) { continue; }

		h_min = ntohl( my_nets[sn]->range_min );
		h_max = ntohl( my_nets[sn]->range_max );
		for( ss=h_min; ss <= h_max; ss++ ) {
			n_ord = htonl( ss );
			if ( my_CheckIP( n_ord ) ) {
				message->lease_type = lease_type;
				message->cpe_type = cpe_type;
				message->ipaddr = n_ord;
				inp.s_addr = n_ord;
				strncpy( message->s_ipaddr, inet_ntoa( inp ), 20 );
				message->b_ipaddr = ntohl( n_ord );
				message->vlan = my_nets[sn]->cmts_vlan;
				message->lease_time = my_nets[sn]->lease_time;
				my_WriteNewLease( message );
				return;
			}
		}
		my_nets[sn]->full_flag = 1;
		Update_Network_Full_Flag( FULLFLAG_seton, my_nets[sn]->netid );
	}

	my_syslog(LOG_WARNING, "NAK -- no IP addresses to give gi %s mac %s",
		message->in_giaddr, message->s_macaddr );

	{ // Use CountUnusedIP to clear any unneeded FullFlags
		int dummy1, dummy2;
		sn = 0;
		while ( sn < num_nets && my_nets[sn]->cmts_ip != message->in_pack.giaddr ) sn++;
		if (sn < num_nets) {
			my_CountUnusedIP( &dummy1, &dummy2, my_nets[sn]->cmts_vlan, message->in_pack.giaddr );
		} else {
			my_syslog(LOG_WARNING, "ERR -- could not find GIADDR in net list. gi %s",
				message->in_giaddr );
		}
	}

	new_time = time( NULL );
	if ( (new_time - old_time) > 60 ) {	/* if there are no IPs available */
		old_time = new_time;		/* reload every minute */

		my_Load_Nets();

		if ( (new_time - old_time2) > 900 ) {  /* send alert every 15 minutes */
			old_time2 = new_time;

			alert_pgm = GetConfigVar( "alert-program" );
			if (*alert_pgm != ' ') {
				char    *argv[4];
				pid_t	new_pid;

				new_pid = fork();
				if (new_pid == 0) {
					argv[0] = alert_pgm;
					argv[1] = "NO_IP";
					argv[2] = message->in_giaddr;
					argv[3] = NULL;
					execvp( alert_pgm, argv );
					fprintf( stderr, "cannot find %s\n", alert_pgm );
					exit(0);
				}
			}
		}
	}
}



int my_CheckIP( u_int32_t ipaddr ) {
	MYSQL_RES	*res;
	MYSQL_ROW	myrow;
	char		qbuf[500], *ip;
	struct in_addr	tmpval;

	tmpval.s_addr = ipaddr;
	ip = inet_ntoa( tmpval );

	strcpy( qbuf, "select macaddr from dhcp_leases where ipaddr = '" );
	strcat( qbuf, ip );
	strcat( qbuf, "' " );

	if ( mysql_ping( my_sock ) ) {
		fprintf( stderr, "no sql server -- %s\n", mysql_error( my_sock ) );
		return 0;
	}
	if ( mysql_query( my_sock, qbuf ) ) {
		fprintf( stderr, "sql error\nq=%s \ne=%s\n", qbuf,
			mysql_error( my_sock ) );
		return 0;
	}
	res = mysql_store_result( my_sock );
	if (res == NULL) return 0;
	myrow = mysql_fetch_row( res );
	if (myrow == NULL) {
		mysql_free_result(res);		/* if no result IP is available */
		return 1;
	} else {
		mysql_free_result(res);		/* IP is in use */
		return 0;
	}
}


void my_Count_Allowed_IPs( dhcp_message *message ) {
	MYSQL_RES	*res;
	MYSQL_ROW	myrow;
	char		qbuf[500];

	strcpy( qbuf, "select subnum, cfg_static_ip, cfg_dynamic_ip " );
	strcat( qbuf, "from docsis_modem where modem_macaddr = '" );
	strcat( qbuf, message->s_modem_macaddr );
	strcat( qbuf, "' " );

	if ( mysql_ping( my_sock ) ) {
		fprintf( stderr, "no sql server -- %s\n", mysql_error( my_sock ) );
		return;
	}
	if ( mysql_query( my_sock, qbuf ) ) {
		fprintf( stderr, "sql error\nq=%s \ne=%s\n", qbuf,
			mysql_error( my_sock ) );
		return;
	}
	res = mysql_store_result( my_sock );
	if (res == NULL) return;
	myrow = mysql_fetch_row( res );
	if (myrow == NULL) { mysql_free_result(res); return; }

	message->subnum = atoll( (myrow)[0] );
	message->static_ips = atoi( (myrow)[1] );
	message->dynamic_ips = atoi( (myrow)[2] );

	mysql_free_result(res);
}

void my_Count_Used_IPs( dhcp_message *message ) {
	MYSQL_RES	*res;
	MYSQL_ROW	myrow;
	char		qbuf[500], qbuf2[500], *use_subnum;

	use_subnum = GetConfigVar( "use-subnum" );
	if (*use_subnum == 'y') {
		strcpy( qbuf, "select ipaddr, dynamic_flag from dhcp_leases " );
		strcat( qbuf, "where modem_macaddr = '%s' and subnum = %ld " );
		snprintf( qbuf2, 500, qbuf, message->s_modem_macaddr, message->subnum );
	} else {
		strcpy( qbuf, "select ipaddr, dynamic_flag from dhcp_leases " );
		strcat( qbuf, "where modem_macaddr = '%s' " );
		snprintf( qbuf2, 500, qbuf, message->s_modem_macaddr );
	}
	if ( mysql_ping( my_sock ) ) {
		fprintf( stderr, "no sql server -- %s\n", mysql_error( my_sock ) );
		return;
	}
	if ( mysql_query( my_sock, qbuf2 ) ) {
		fprintf( stderr, "sql error\nq=%s \ne=%s\n", qbuf,
			mysql_error( my_sock ) );
		return;
	}
	res = mysql_store_result( my_sock );
	if (res == NULL) return;
	while ( (myrow = mysql_fetch_row( res )) ) {
		if ( strcmp((myrow)[1],"YES") == 0 ) {
			message->dynamic_cur++;
		} else {
			message->static_cur++;
		}
	}
	mysql_free_result(res);
}

/* ****************************************************************** */


void  my_WriteNewLease( dhcp_message *message ) {
	char	*s_dynam;
	char	qbuf1[500];
	char	qbuf2[500];

	s_dynam = "NO";
	if (message->cpe_type == CPE_DYNAMIC) { s_dynam = "YES"; }

	strcpy( qbuf1, "replace into dhcp_leases " );
	strcat( qbuf1, "( ipaddr, macaddr, start_time, end_time, update_time, " );
	strcat( qbuf1, "cmts_vlan, dynamic_flag, subnum ) " );
	strcat( qbuf1, " values ( '%s', '%s', now(), now(), now(), %d, '%s', %llu ) " );

	snprintf( qbuf2, 500, qbuf1, message->s_ipaddr, message->s_macaddr, message->vlan,
				s_dynam, message->subnum );

	if ( mysql_ping( my_sock ) ) {
		fprintf( stderr, "no sql server -- %s\n", mysql_error( my_sock ) );
		return;
	}
	if ( mysql_query( my_sock, qbuf2 ) ) {
		fprintf( stderr, "sql error\nq=%s \ne=%s\n", qbuf2,
			mysql_error( my_sock ) );
	}
}


/* ****************************************************************** */


void  my_UpdateLease( dhcp_message *message ) {
	char	qbuf1[1500];
	char	qbuf2[1500];

	strcpy( qbuf1, "delete from dhcp_leases ");
	strcat( qbuf1, "where macaddr = '%s-x' and lockip_flag = 'NO' ");
	sprintf( qbuf2, qbuf1, message->s_macaddr );
	if ( mysql_ping( my_sock ) ) {
		fprintf( stderr, "no sql server -- %s\n", mysql_error( my_sock ) );
		return;
	}
	if ( mysql_query( my_sock, qbuf2 ) ) {
		fprintf( stderr, "sql error\nq=%s \ne=%s\n", qbuf2,
			mysql_error( my_sock ) );
	}

	if (message->in_opts.agent_mac_len > 0) {
		strcpy( qbuf1, "update dhcp_leases " );
		strcat( qbuf1, "set modem_macaddr = '%s' where ipaddr = '%s' " );

		sprintf( qbuf2, qbuf1, message->s_modem_macaddr, message->s_ipaddr );

		if ( mysql_ping( my_sock ) ) {
			fprintf( stderr, "no sql server -- %s\n", mysql_error( my_sock ) );
			return;
		}
		if ( mysql_query( my_sock, qbuf2 ) ) {
			fprintf( stderr, "sql error\nq=%s \ne=%s\n", qbuf2,
				mysql_error( my_sock ) );
		}
	}

	if (message->in_opts.host_name_len > 0) {
		strcpy( qbuf1, "update dhcp_leases " );
		strcat( qbuf1, "set pc_name = '%s' where ipaddr = '%s' " );

		sprintf( qbuf2, qbuf1, message->in_opts.host_name, message->s_ipaddr );

		if ( mysql_ping( my_sock ) ) {
			fprintf( stderr, "no sql server -- %s\n", mysql_error( my_sock ) );
			return;
		}
		if ( mysql_query( my_sock, qbuf2 ) ) {
			fprintf( stderr, "sql error\nq=%s \ne=%s\n", qbuf2,
				mysql_error( my_sock ) );
		}
	}

	if (message->subnum > 0) {
		strcpy( qbuf1, "update dhcp_leases " );
		strcat( qbuf1, "set subnum = '%ld' where ipaddr = '%s' " );

		sprintf( qbuf2, qbuf1, message->subnum, message->s_ipaddr );

		if ( mysql_ping( my_sock ) ) {
			fprintf( stderr, "no sql server -- %s\n", mysql_error( my_sock ) );
			return;
		}
		if ( mysql_query( my_sock, qbuf2 ) ) {
			fprintf( stderr, "sql error\nq=%s \ne=%s\n", qbuf2,
				mysql_error( my_sock ) );
		}
	}

	strcpy( qbuf1, "update dhcp_leases " );
	strcat( qbuf1, "set start_time = now(), ");
	strcat( qbuf1, "end_time = date_add(now(),interval %ld second), " );
	strcat( qbuf1, "cmts_vlan = %d ");
	strcat( qbuf1, "where ipaddr = '%s' " );

	sprintf( qbuf2, qbuf1, message->lease_time, message->vlan, message->s_ipaddr );

	if ( mysql_ping( my_sock ) ) {
		fprintf( stderr, "no sql server -- %s\n", mysql_error( my_sock ) );
		return;
	}
	if ( mysql_query( my_sock, qbuf2 ) ) {
		fprintf( stderr, "sql error\nq=%s \ne=%s\n", qbuf2,
			mysql_error( my_sock ) );
	} else {
		// fprintf( stderr, "why-- %s\n", qbuf2 );
	}
}


/* ****************************************************************** */


void  my_UpdateTime( dhcp_message *message ) {
	char	qbuf1[1500];
	char	qbuf2[1500];

	strcpy( qbuf1, "update dhcp_leases " );
	strcat( qbuf1, "set update_time = now() " );
	strcat( qbuf1, "where ipaddr = '%s' " );
	sprintf( qbuf2, qbuf1, message->s_ipaddr );

	if ( mysql_ping( my_sock ) ) {
		fprintf( stderr, "no sql server -- %s\n", mysql_error( my_sock ) );
		return;
	}
	if ( mysql_query( my_sock, qbuf2 ) ) {
		fprintf( stderr, "sql error\nq=%s \ne=%s\n", qbuf2,
			mysql_error( my_sock ) );
	} else {
		// fprintf(stderr, "updatetime '%s'\n", qbuf2 );
	}
}


/* ****************************************************************** */

void  my_DeleteLease( dhcp_message *message ) {
	char    qbuf1[1500];
	char    qbuf2[1500];

	strcpy( qbuf1, "insert into dhcp_oldleases ");
	strcat( qbuf1, "select now(), ipaddr, macaddr, start_time, end_time, ");
	strcat( qbuf1, "update_time, dynamic_flag, pc_name, subnum, modem_macaddr, cmts_vlan " );
	strcat( qbuf1, "from dhcp_leases " );
	strcat( qbuf1, "where ipaddr = '%s' and lockip_flag = 'NO' " );
	sprintf( qbuf2, qbuf1, message->s_ipaddr );

	if ( mysql_ping( my_sock ) ) {
		fprintf( stderr, "no sql server -- %s\n", mysql_error( my_sock ) );
		return;
	}
	if ( mysql_query( my_sock, qbuf2 ) ) {
		fprintf( stderr, "sql error\nq=%s \ne=%s\n", qbuf2,
			mysql_error( my_sock ) );
		return;
	}

	strcpy( qbuf1, "update dhcp_leases set macaddr = '%s-x', ");
	strcat( qbuf1, "modem_macaddr = 'delete' ");
	strcat( qbuf1, "where ipaddr = '%s' and lockip_flag = 'NO' " );
	sprintf( qbuf2, qbuf1, message->s_macaddr, message->s_ipaddr );

	// strcpy( qbuf1, "delete from dhcp_leases ");
	// strcat( qbuf1, "where ipaddr = '%s' " );
	// sprintf( qbuf2, qbuf1, message->s_ipaddr );

	if ( mysql_ping( my_sock ) ) {
		fprintf( stderr, "no sql server -- %s\n", mysql_error( my_sock ) );
		return;
	}
	if ( mysql_query( my_sock, qbuf2 ) ) {
		fprintf( stderr, "sql error\nq=%s \ne=%s\n", qbuf2,
			mysql_error( my_sock ) );
	}

	strcpy( message->s_ipaddr, "" );
	message->b_ipaddr = 0;
	message->ipaddr = 0;
	message->vlan = 0;
}


/* ****************************************************************** */


/* include SQL functions for the DHCP lease functions */
/* functions to count leases and delete leases */
/*	void  my_CountUnusedIP( int *unused, int *total );
	void my_CountOLDIP( int *ip45, int *ip30, int *ip15 );
	void my_DeleteOldestLease( void );
*/
#include "my_sql_misc_lease.1"


/* include SQL functions for the DHCP agent functions */
/*	void  my_UpdateAgent( dhcp_message *message );
	int   my_findAgent( dhcp_message *message );
	int   my_findAgent_QUERY( char *qbuf, dhcp_message *message );
	int   my_findAgent_QUERY2( char *qbuf, dhcp_message *message );
*/
#include "my_sql_modem_update.1"


/* include Cache functions for the TFTP server */
/*	void my_CONFIG_Load( void );
	int  add_Config_Item( u_int8_t *dst, u_int32_t idnum, u_int32_t maxsize );
	void add_MIC_Item( u_int32_t *cm_mic, u_int32_t *cmts_mic );
	int  Check_MIC_Item( u_int32_t *cm_mic, u_int32_t *cmts_mic );
*/
#include "my_sql_tftp_filegen.1"


/* include SQL functions for the TFTP server */
/*	void my_TFTPlog( char * fname, u_int32_t ipaddr );
*/
#include "my_sql_tftp_log.1"


/* include SQL functions for the SYSLOG server */
/*	void my_LogSyslog( u_int32_t ipaddr, char * message );
*/
#include "my_sql_syslog.1"


/* include MAC address conversion utilities */
/*	void cv_mac2num( u_int64_t *b_mac, u_int8_t *chaddr );
	void cv_macaddr( u_int8_t *chaddr, char *macaddr, int numbytes );
	void cv_addrmac( u_int8_t *chaddr, char *macaddr, int numbytes );
*/
#include "my_sql_conversion.1"


/* include Log summerize routines */
	void  Summerize_Logs( char *qbuf );
	void  Delete_Logs( char *qbuf );
	void  Log_Summerize( void );
#include "my_sql_summerize.1"

