/* super_server.c
 *
 * docsis_server
 *
 * Copyright (C) 2002 Access Communications
 *                    email docsis_guy@accesscomm.ca
 *
 * Designed for use with DOCSIS cable modems and hosts.
 * And for use with MySQL Database Server.
 *
 ********************************************************************
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <pwd.h>
#include <string.h>
#include <signal.h>
#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>

#include <sys/time.h>
#include <sys/resource.h>

#include "src-config/config_file.h"
#include "src-mysql/my_sql.h"
#include "Version.h"


void my_exit( void );

void dhcp_server( void );
void tftp_server( void );
void time_server( void );
void syslog_server( void );
void cf_generator( void );
void backgrounder( void );


int main( int argc, char *argv[] ) {
	pid_t		pid[6];
	int		i, rename = 0;
	char		*dt, *tim;
	struct rlimit	rlim;

	setsid();	/* Set Process Independency */
	umask(027);	/* Set File Creation Mask */
	chdir( CHROOT_PATH );	/* Change to /tmp dir */

	rlim.rlim_cur = RLIM_INFINITY;
	rlim.rlim_max = RLIM_INFINITY;
	setrlimit(RLIMIT_CORE, &rlim );

	/* Lets Fork Really Early. */
	pid[0] = fork();
	if (pid[0] != 0) { exit(0); }
	pid[0] = fork();
	if (pid[0] != 0) { exit(0); }

	for (i=getdtablesize();i>=0;--i) close(i); /* close all descriptors */
	i=open( CHROOT_PATH "/" LOGFILE, O_RDWR | O_CREAT | O_APPEND, S_IRUSR | S_IWUSR | S_IRGRP );
	dup(i);
	dup(i);

	signal( SIGINT,  (sig_t) my_exit );
	signal( SIGQUIT, (sig_t) my_exit );
	signal( SIGTERM, (sig_t) my_exit );

	if (argc > 1) {
		if (strlen(argv[1]) > 10) {
			rename = strlen(argv[1]);
			bzero( argv[1], rename );
		}
	}
	for (i=0;i<6;i++) pid[i] = 0;

	ReadConfigFile();

	get_date_time( &dt, &tim );
	fprintf(stderr, "Starting Version %s  ", VERSION );
	fprintf(stderr, "Date: %s  Time: %s \n", dt, tim );

	if (my_CheckActiveService("dhcp")) {
		fprintf(stderr, "Now Forking DHCP Server." );
		pid[0] = fork();
		if (pid[0] == 0) {
			if (rename) { strcpy( argv[0], "docsis_server (DHCP)     " ); }
			dhcp_server();
			exit(0);
		} else {
			fprintf(stderr, " PID = %d \n", pid[0] );
		}
	}

	if (my_CheckActiveService("tftp")) {
		fprintf(stderr, "Now Forking TFTP Server." );
		pid[1] = fork();
		if (pid[1] == 0) {
			if (rename) { strcpy( argv[0], "docsis_server (TFTP)    " ); }
			tftp_server();
			exit(0);
		} else {
			fprintf(stderr, " PID = %d \n", pid[1] );
		}
	}

	if (my_CheckActiveService("time")) {
		fprintf(stderr, "Now Forking TIME Server." );
		pid[2] = fork();
		if (pid[2] == 0) {
			if (rename) { strcpy( argv[0], "docsis_server (TIME)    " ); }
			time_server();
			exit(0);
		} else {
			fprintf(stderr, " PID = %d \n", pid[2] );
		}
	}


	if (my_CheckActiveService("syslog")) {
		fprintf(stderr, "Now Forking SYSLOG Server." );
		pid[3] = fork();
		if (pid[3] == 0) {
			if (rename) { strcpy( argv[0], "docsis_server (SYSLOG)   " ); }
			syslog_server();
			exit(0);
		} else {
			fprintf(stderr, " PID = %d \n", pid[3] );
		}
	}

	if (my_CheckActiveService("backg")) {
		fprintf(stderr, "Now Forking BACKG Server." );
		pid[4] = fork();
		if (pid[4] == 0) {
			if (rename) { strcpy( argv[0], "docsis_server (BACKG)   " ); }
			backgrounder();
			exit(0);
		} else {
			fprintf(stderr, " PID = %d \n", pid[4] );
		}
	}

	{
		FILE	*pidfile = NULL;
		pidfile = fopen( "/var/run/docsis_server.pid", "w" );
		if (pidfile != NULL) {
			if (pid[0]) fprintf(pidfile, "%d",  pid[0] );
			if (pid[1]) fprintf(pidfile, " %d", pid[1] );
			if (pid[2]) fprintf(pidfile, " %d", pid[2] );
			if (pid[3]) fprintf(pidfile, " %d", pid[3] );
			if (pid[4]) fprintf(pidfile, " %d", pid[4] );
			fprintf(pidfile, "\n" );
			fclose( pidfile );
		}
	}

	exit(0);
}


void my_exit( void ) {
	my_flush_SQL();
	exit(0);
}


