#include <stdio.h>
#include <unistd.h>
#include <stdlib.h>
#include <string.h>

#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>

#include "../src-dhcpd/dhcpd.h"
#include "../src-mysql/my_sql.h"
#include "../src-config/config_file.h"


typedef struct _net_count {
	u_int32_t	cmts_ip;
	int		cnt;

	struct _net_count *next;
} net_count;


extern int            num_nets;
extern ConfigNetsPtr  my_nets[];


int main( int argc, char *argv[] ) {
	int	unused, total, vlan_id;
	int	old45, old30, old15;
	float	percent_unused;
	net_count	counter, *x = NULL;
	struct in_addr sin;
	int	sn;

	if (argc < 2) {
		printf( "usage: %s vlan_id\n", argv[0] );
		exit(1);
	}

	vlan_id = atoi( argv[1] );

	ReadConfigFile();
	InitSQLconnection();
	my_Load_Nets();

	counter.cmts_ip = 0;
	counter.cnt = 0;
	counter.next = NULL;

	for( sn=0; sn < num_nets; sn++ ) {
		if (my_nets[sn]->nettype == NET_TYPE_CPE &&
		    my_nets[sn]->grant_flag == 1 &&
		    my_nets[sn]->cmts_vlan == vlan_id) {
			x = &counter;
			while( x != NULL ) {
				if (x->cmts_ip == my_nets[sn]->cmts_ip) {
					x->cnt++;
					x = NULL;
				} else {
					if (x->next == NULL) {
						x->next = malloc(sizeof(net_count));
						x->next->cmts_ip = my_nets[sn]->cmts_ip;
						x->next->cnt = 0;
						x->next->next = NULL;
					}
					x = x->next;
				}
			}
		}
	}

	x = &counter;
	while( x != NULL ) {
		if (x->cmts_ip == 0) { x = x->next; continue; }

		sin.s_addr = x->cmts_ip;
		printf( "VLAN %d  --  CMTS IP %s\n", vlan_id, inet_ntoa( sin ) );

		my_CountUnusedIP( &unused, &total, vlan_id, x->cmts_ip );
		percent_unused = ( (float) unused / (float) total ) * 100;

		printf( "  Total IP Addresses = %d\n", total );
		printf( "  IP Addresses Used = %d      %2.2f\n", (total - unused),
			(100.0 - percent_unused) );
		printf( "  IP Addresses Unused = %d    %2.2f\n", unused, percent_unused );

		x = x->next;
	}
	
	my_CountOLDIP( vlan_id, &old45, &old30, &old15 );

	printf( "\n\n" );
	printf( " IPs stale for 45 days = %d\n", old45 );
	printf( " Actual IP Used = %d\n", (total - unused - old45) );
	printf( " Actual IP Unused = %d\n", (unused + old45) );

	printf( "\n" );
	printf( " IPs stale for 30 days = %d\n", old30 );
	printf( " Actual IP Used = %d\n", (total - unused - old30) );
	printf( " Actual IP Unused = %d\n", (unused + old30) );

	printf( "\n" );
	printf( " IPs stale for 15 days = %d\n", old15 );
	printf( " Actual IP Used = %d\n", (total - unused - old15) );
	printf( " Actual IP Unused = %d\n", (unused + old15) );

	exit(0);
}
